import json

from moto.core.responses import BaseResponse

from .models import BudgetsBackend, budgets_backends


class BudgetsResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="budgets")

    @property
    def backend(self) -> BudgetsBackend:
        return budgets_backends[self.current_account][self.partition]

    def create_budget(self) -> str:
        account_id = self._get_param("AccountId")
        budget = self._get_param("Budget")
        notifications = self._get_param("NotificationsWithSubscribers", [])
        self.backend.create_budget(
            account_id=account_id, budget=budget, notifications=notifications
        )
        return json.dumps(dict())

    def describe_budget(self) -> str:
        account_id = self._get_param("AccountId")
        budget_name = self._get_param("BudgetName")
        budget = self.backend.describe_budget(
            account_id=account_id, budget_name=budget_name
        )
        return json.dumps(dict(Budget=budget))

    def describe_budgets(self) -> str:
        account_id = self._get_param("AccountId")
        budgets = self.backend.describe_budgets(account_id=account_id)
        return json.dumps(dict(Budgets=budgets, nextToken=None))

    def delete_budget(self) -> str:
        account_id = self._get_param("AccountId")
        budget_name = self._get_param("BudgetName")
        self.backend.delete_budget(account_id=account_id, budget_name=budget_name)
        return json.dumps(dict())

    def create_notification(self) -> str:
        account_id = self._get_param("AccountId")
        budget_name = self._get_param("BudgetName")
        notification = self._get_param("Notification")
        subscribers = self._get_param("Subscribers")
        self.backend.create_notification(
            account_id=account_id,
            budget_name=budget_name,
            notification=notification,
            subscribers=subscribers,
        )
        return json.dumps(dict())

    def delete_notification(self) -> str:
        account_id = self._get_param("AccountId")
        budget_name = self._get_param("BudgetName")
        notification = self._get_param("Notification")
        self.backend.delete_notification(
            account_id=account_id, budget_name=budget_name, notification=notification
        )
        return json.dumps(dict())

    def describe_notifications_for_budget(self) -> str:
        account_id = self._get_param("AccountId")
        budget_name = self._get_param("BudgetName")
        notifications = self.backend.describe_notifications_for_budget(
            account_id=account_id, budget_name=budget_name
        )
        return json.dumps(dict(Notifications=notifications, NextToken=None))



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_create_configuration_profile():
    client = boto3.client("appconfig", region_name="eu-north-1")
    app_id = client.create_application(Name="testapp")["Id"]
    resp = client.create_configuration_profile(
        ApplicationId=app_id,
        Name="config_name",
        Description="desc",
        LocationUri="luri",
        RetrievalRoleArn="rrarn:rrarn:rrarn:rrarn",
        Validators=[{"Type": "JSON", "Content": "c"}],
        Type="freeform",
    )
    del resp["ResponseMetadata"]
    assert "Id" in resp
    config_profile_id = resp.pop("Id")

    expected_response = {
        "ApplicationId": app_id,
        "Name": "config_name",
        "Description": "desc",
        "LocationUri": "luri",
        "RetrievalRoleArn": "rrarn:rrarn:rrarn:rrarn",
        "Validators": [{"Type": "JSON", "Content": "c"}],
        "Type": "freeform",
    }

    assert resp == expected_response

    resp = client.get_configuration_profile(
        ApplicationId=app_id,
        ConfigurationProfileId=config_profile_id,
    )
    del resp["ResponseMetadata"]
    assert "Id" in resp
    resp.pop("Id")
    assert resp == expected_response

    profiles = client.list_configuration_profiles(ApplicationId=app_id)["Items"]
    assert profiles == [
        {
            "ApplicationId": app_id,
            "Id": config_profile_id,
            "Name": "config_name",
            "LocationUri": "luri",
            "Type": "freeform",
        }
    ]

    update = client.update_configuration_profile(
        ApplicationId=app_id,
        ConfigurationProfileId=config_profile_id,
        Name="name2",
        Description="desc2",
        RetrievalRoleArn="rrarn:rrarn:rrarn:222",
        Validators=[],
    )
    assert update["Name"] == "name2"
    assert update["RetrievalRoleArn"] == "rrarn:rrarn:rrarn:222"

    client.delete_configuration_profile(
        ApplicationId=app_id,
        ConfigurationProfileId=config_profile_id,
    )

    with pytest.raises(ClientError) as exc:
        client.get_configuration_profile(
            ApplicationId=app_id,
            ConfigurationProfileId=config_profile_id,
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_tag_config_profile():
    client = boto3.client("appconfig", region_name="us-east-2")
    app_id = client.create_application(Name="testapp")["Id"]
    cp_id = client.create_configuration_profile(
        ApplicationId=app_id,
        Name="config_name",
        LocationUri="luri",
    )["Id"]
    cp_arn = f"arn:aws:appconfig:us-east-2:123456789012:application/{app_id}/configurationprofile/{cp_id}"

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {}

    client.tag_resource(
        ResourceArn=cp_arn,
        Tags={"k1": "v1"},
    )

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {"k1": "v1"}

    ####
    # Check this flow works when creating an app with tags
    app_id = client.create_application(Name="testapp")["Id"]
    cp_id = client.create_configuration_profile(
        ApplicationId=app_id,
        Name="config_name",
        LocationUri="luri",
        Tags={"k1": "v1"},
    )["Id"]
    cp_arn = f"arn:aws:appconfig:us-east-2:123456789012:application/{app_id}/configurationprofile/{cp_id}"

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {"k1": "v1"}

    client.tag_resource(
        ResourceArn=cp_arn,
        Tags={"k2": "v2", "k3": "v3"},
    )

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {"k1": "v1", "k2": "v2", "k3": "v3"}

    client.untag_resource(ResourceArn=cp_arn, TagKeys=["k2"])

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {"k1": "v1", "k3": "v3"}

    client.untag_resource(ResourceArn=cp_arn, TagKeys=["k1", "k3"])

    tags = client.list_tags_for_resource(ResourceArn=cp_arn)["Tags"]
    assert tags == {}



"""Exceptions raised by the Directory Service service."""

from typing import List, Tuple

from moto.core.exceptions import JsonRESTError


class DsValidationException(JsonRESTError):
    """Report one of more parameter validation errors."""

    code = 400

    def __init__(self, error_tuples: List[Tuple[str, str, str]]):
        """Validation errors are concatenated into one exception message.

        error_tuples is a list of tuples.  Each tuple contains:

          - name of invalid parameter,
          - value of invalid parameter,
          - string describing the constraints for that parameter.
        """
        msg_leader = (
            f"{len(error_tuples)} "
            f"validation error{'s' if len(error_tuples) > 1 else ''} detected: "
        )
        msgs = []
        for arg_name, arg_value, constraint in error_tuples:
            value = "at" if "assword" in arg_name else f"'{arg_value}' at"
            msgs.append(
                f"Value {value} '{arg_name}' failed to satisfy constraint: "
                f"Member must {constraint}"
            )
        super().__init__("ValidationException", msg_leader + "; ".join(msgs))


class ClientException(JsonRESTError):
    """Client exception has occurred. VPC parameters are invalid."""

    code = 400

    def __init__(self, message: str):
        super().__init__("ClientException", message)


class DirectoryLimitExceededException(JsonRESTError):
    """Maximum number of directories in region has been reached."""

    code = 400

    def __init__(self, message: str):
        super().__init__("DirectoryLimitExceededException", message)


class EntityDoesNotExistException(JsonRESTError):
    """The specified entity could not be found."""

    code = 400

    def __init__(self, message: str):
        super().__init__("EntityDoesNotExistException", message)


class EntityAlreadyExistsException(JsonRESTError):
    """The specified entity already exists."""

    code = 400

    def __init__(self, message: str):
        super().__init__("EntityAlreadyExistsException", message)


class InvalidNextTokenException(JsonRESTError):
    """Invalid next token parameter used to return a list of entities."""

    code = 400

    def __init__(self) -> None:
        super().__init__(
            "InvalidNextTokenException",
            "Invalid value passed for the NextToken parameter",
        )


class InvalidParameterException(JsonRESTError):
    """Invalid parameter."""

    code = 400

    def __init__(self, message: str):
        super().__init__("InvalidParameterException", message)


class TagLimitExceededException(JsonRESTError):
    """The maximum allowed number of tags was exceeded."""

    code = 400

    def __init__(self, message: str):
        super().__init__("TagLimitExceededException", message)


class ValidationException(JsonRESTError):
    """Tag validation failed."""

    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class UnsupportedOperationException(JsonRESTError):
    """Unsupported Operation"""

    code = 400

    def __init__(self, message: str):
        super().__init__("UnsupportedOperationException", message)



import json

import boto3
import pytest

from moto import mock_aws


@mock_aws
def test_sns_topic():
    dummy_template = get_template(with_properties=True)
    template_json = json.dumps(dummy_template)
    cf = boto3.client("cloudformation", region_name="us-west-1")
    cf.create_stack(StackName="test_stack", TemplateBody=template_json)

    sns = boto3.client("sns", region_name="us-west-1")
    topics = sns.list_topics()["Topics"]
    assert len(topics) == 1
    topic_arn = topics[0]["TopicArn"]
    assert "my_topics" in topic_arn

    subscriptions = sns.list_subscriptions()["Subscriptions"]
    assert len(subscriptions) == 1
    subscription = subscriptions[0]
    assert subscription["TopicArn"] == topic_arn
    assert subscription["Protocol"] == "https"
    assert topic_arn in subscription["SubscriptionArn"]
    assert subscription["Endpoint"] == "https://example.com"

    stack = cf.describe_stacks(StackName="test_stack")["Stacks"][0]
    topic_name_output = [x for x in stack["Outputs"] if x["OutputKey"] == "topic_name"][
        0
    ]
    assert topic_name_output["OutputValue"] == "my_topics"
    topic_arn_output = [x for x in stack["Outputs"] if x["OutputKey"] == "topic_arn"][0]
    assert topic_arn_output["OutputValue"] == topic_arn


@mock_aws
def test_sns_update_topic():
    dummy_template = get_template(with_properties=True)
    sns_template_json = json.dumps(dummy_template)
    cf = boto3.client("cloudformation", region_name="us-west-1")
    cf.create_stack(StackName="test_stack", TemplateBody=sns_template_json)

    sns = boto3.client("sns", region_name="us-west-1")
    topics = sns.list_topics()["Topics"]
    assert len(topics) == 1

    dummy_template["Resources"]["MySNSTopic"]["Properties"]["Subscription"][0][
        "Endpoint"
    ] = "https://example-updated.com"
    sns_template_json = json.dumps(dummy_template)
    cf.update_stack(StackName="test_stack", TemplateBody=sns_template_json)

    topics = sns.list_topics()["Topics"]
    assert len(topics) == 1
    topic_arn = topics[0]["TopicArn"]
    assert "my_topics" in topic_arn

    subscriptions = sns.list_subscriptions()["Subscriptions"]
    assert len(subscriptions) == 1
    subscription = subscriptions[0]
    assert subscription["TopicArn"] == topic_arn
    assert subscription["Protocol"] == "https"
    assert topic_arn in subscription["SubscriptionArn"]
    assert subscription["Endpoint"] == "https://example-updated.com"


@mock_aws
@pytest.mark.parametrize("with_properties", [True, False])
def test_sns_update_remove_topic(with_properties):
    dummy_template = get_template(with_properties)
    sns_template_json = json.dumps(dummy_template)
    cf = boto3.client("cloudformation", region_name="us-west-1")
    cf.create_stack(StackName="test_stack", TemplateBody=sns_template_json)

    sns = boto3.client("sns", region_name="us-west-1")
    topics = sns.list_topics()["Topics"]
    assert len(topics) == 1

    dummy_template["Resources"].pop("MySNSTopic")
    dummy_template.pop("Outputs")
    sns_template_json = json.dumps(dummy_template)
    cf.update_stack(StackName="test_stack", TemplateBody=sns_template_json)

    topics = sns.list_topics()["Topics"]
    assert len(topics) == 0


@mock_aws
@pytest.mark.parametrize("with_properties", [True, False])
def test_sns_delete_topic(with_properties):
    sns_template_json = json.dumps(get_template(with_properties))
    cf = boto3.client("cloudformation", region_name="us-west-1")
    cf.create_stack(StackName="test_stack", TemplateBody=sns_template_json)

    sns = boto3.client("sns", region_name="us-west-1")
    topics = sns.list_topics()["Topics"]
    assert len(topics) == 1

    cf.delete_stack(StackName="test_stack")

    topics = sns.list_topics()["Topics"]
    assert len(topics) == 0


def get_template(with_properties):
    dummy_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {"MySNSTopic": {"Type": "AWS::SNS::Topic"}},
        "Outputs": {
            "topic_name": {"Value": {"Fn::GetAtt": ["MySNSTopic", "TopicName"]}},
            "topic_arn": {"Value": {"Ref": "MySNSTopic"}},
        },
    }
    if with_properties:
        dummy_template["Resources"]["MySNSTopic"]["Properties"] = {
            "Subscription": [{"Endpoint": "https://example.com", "Protocol": "https"}],
            "TopicName": "my_topics",
        }
    return dummy_template



"""CloudHSMV2Backend class with methods for supported APIs."""

import uuid
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.utils import utcnow
from moto.utilities.paginator import paginate

from .exceptions import ResourceNotFoundException


class Cluster:
    def __init__(
        self,
        backup_retention_policy: Optional[Dict[str, str]],
        hsm_type: str,
        source_backup_id: Optional[str],
        subnet_ids: List[str],
        network_type: str = "IPV4",
        tag_list: Optional[List[Dict[str, str]]] = None,
        mode: str = "DEFAULT",
        region_name: str = "us-east-1",
    ):
        self.cluster_id = str(uuid.uuid4())
        self.backup_policy = "DEFAULT"
        self.backup_retention_policy = backup_retention_policy
        self.create_timestamp = utcnow()
        self.hsms: List[Dict[str, Any]] = []
        self.hsm_type = hsm_type
        self.source_backup_id = source_backup_id
        self.state = "ACTIVE"
        self.state_message = "The cluster is ready for use."
        self.subnet_mapping = {subnet_id: region_name for subnet_id in subnet_ids}
        self.vpc_id = "vpc-" + str(uuid.uuid4())[:8]
        self.network_type = network_type
        self.certificates = {
            "ClusterCsr": "",
            "HsmCertificate": "",
            "AwsHardwareCertificate": "",
            "ManufacturerHardwareCertificate": "",
            "ClusterCertificate": "",
        }
        self.tag_list = tag_list or []
        self.mode = mode

    def to_dict(self) -> Dict[str, Any]:
        return {
            "BackupPolicy": self.backup_policy,
            "BackupRetentionPolicy": self.backup_retention_policy,
            "ClusterId": self.cluster_id,
            "CreateTimestamp": self.create_timestamp,
            "Hsms": self.hsms,
            "HsmType": self.hsm_type,
            "SourceBackupId": self.source_backup_id,
            "State": self.state,
            "StateMessage": self.state_message,
            "SubnetMapping": self.subnet_mapping,
            "VpcId": self.vpc_id,
            "NetworkType": self.network_type,
            "Certificates": self.certificates,
            "TagList": self.tag_list,
            "Mode": self.mode,
        }


class Backup:
    def __init__(
        self,
        cluster_id: str,
        hsm_type: str,
        mode: str,
        tag_list: Optional[List[Dict[str, str]]],
        source_backup: Optional[str] = None,
        source_cluster: Optional[str] = None,
        source_region: Optional[str] = None,
        never_expires: bool = False,
        region_name: str = "us-east-1",
    ):
        self.backup_id = str(uuid.uuid4())
        self.backup_arn = (
            f"arn:aws:cloudhsm:{region_name}:123456789012:backup/{self.backup_id}"
        )
        self.backup_state = "READY"
        self.cluster_id = cluster_id
        self.create_timestamp = utcnow()
        self.copy_timestamp = utcnow() if source_backup else None
        self.never_expires = never_expires
        self.source_region = source_region
        self.source_backup = source_backup
        self.source_cluster = source_cluster
        self.delete_timestamp = None
        self.tag_list = tag_list or []
        self.hsm_type = hsm_type
        self.mode = mode

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "BackupId": self.backup_id,
            "BackupArn": self.backup_arn,
            "BackupState": self.backup_state,
            "ClusterId": self.cluster_id,
            "CreateTimestamp": self.create_timestamp,
            "NeverExpires": self.never_expires,
            "TagList": self.tag_list,
            "HsmType": self.hsm_type,
            "Mode": self.mode,
        }

        if self.copy_timestamp:
            result["CopyTimestamp"] = self.copy_timestamp
        if self.source_region:
            result["SourceRegion"] = self.source_region
        if self.source_backup:
            result["SourceBackup"] = self.source_backup
        if self.source_cluster:
            result["SourceCluster"] = self.source_cluster
        if self.delete_timestamp:
            result["DeleteTimestamp"] = self.delete_timestamp

        return result


class CloudHSMV2Backend(BaseBackend):
    """Implementation of CloudHSMV2 APIs."""

    PAGINATION_MODEL = {
        "describe_backups": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "backup_id",
            "fail_on_invalid_token": False,
        },
        "describe_clusters": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "ClusterId",
            "fail_on_invalid_token": False,
        },
    }

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.tags: Dict[str, List[Dict[str, str]]] = {}
        self.clusters: Dict[str, Cluster] = {}
        self.resource_policies: Dict[str, str] = {}
        self.backups: Dict[str, Backup] = {}

    def list_tags(
        self, resource_id: str, next_token: str, max_results: int
    ) -> Tuple[List[Dict[str, str]], Optional[str]]:
        """
        Pagination is not yet implemented
        """
        if resource_id not in self.tags:
            return [], None

        tags = sorted(self.tags.get(resource_id, []), key=lambda x: x["Key"])

        return tags, None

    def tag_resource(
        self, resource_id: str, tag_list: List[Dict[str, str]]
    ) -> Dict[str, Any]:
        if resource_id not in self.tags:
            self.tags[resource_id] = []

        for new_tag in tag_list:
            tag_exists = False
            for existing_tag in self.tags[resource_id]:
                if existing_tag["Key"] == new_tag["Key"]:
                    existing_tag["Value"] = new_tag["Value"]
                    tag_exists = True
                    break
            if not tag_exists:
                self.tags[resource_id].append(new_tag)

        return {}

    def untag_resource(
        self, resource_id: str, tag_key_list: List[str]
    ) -> Dict[str, Any]:
        if resource_id in self.tags:
            self.tags[resource_id] = [
                tag for tag in self.tags[resource_id] if tag["Key"] not in tag_key_list
            ]

        return {}

    def create_cluster(
        self,
        backup_retention_policy: Optional[Dict[str, str]],
        hsm_type: str,
        source_backup_id: Optional[str],
        subnet_ids: List[str],
        network_type: Optional[str],
        tag_list: Optional[List[Dict[str, str]]],
        mode: Optional[str],
    ) -> Dict[str, Any]:
        cluster = Cluster(
            backup_retention_policy=backup_retention_policy,
            hsm_type=hsm_type,
            source_backup_id=source_backup_id,
            subnet_ids=subnet_ids,
            network_type=network_type or "IPV4",
            tag_list=tag_list,
            mode=mode or "DEFAULT",
            region_name=self.region_name,
        )
        self.clusters[cluster.cluster_id] = cluster

        backup = Backup(
            cluster_id=cluster.cluster_id,
            hsm_type=hsm_type,
            mode=mode or "DEFAULT",
            tag_list=tag_list,
            region_name=self.region_name,
        )
        self.backups[backup.backup_id] = backup

        return cluster.to_dict()

    def delete_cluster(self, cluster_id: str) -> Dict[str, Any]:
        if cluster_id not in self.clusters:
            raise ResourceNotFoundException(f"Cluster {cluster_id} not found")

        cluster = self.clusters[cluster_id]
        cluster.state = "DELETED"
        cluster.state_message = "Cluster deleted"
        del self.clusters[cluster_id]
        return cluster.to_dict()

    @paginate(pagination_model=PAGINATION_MODEL)
    def describe_clusters(
        self, filters: Optional[Dict[str, List[str]]] = None
    ) -> List[Dict[str, str]]:
        clusters = list(self.clusters.values())

        if filters:
            for key, values in filters.items():
                if key == "clusterIds":
                    clusters = [c for c in clusters if c.cluster_id in values]
                elif key == "states":
                    clusters = [c for c in clusters if c.state in values]
                elif key == "vpcIds":
                    clusters = [c for c in clusters if c.vpc_id in values]

        clusters = sorted(clusters, key=lambda x: x.create_timestamp)
        return [c.to_dict() for c in clusters]

    def get_resource_policy(self, resource_arn: str) -> Optional[str]:
        return self.resource_policies.get(resource_arn)

    @paginate(PAGINATION_MODEL)
    def describe_backups(
        self,
        filters: Optional[Dict[str, List[str]]],
        shared: Optional[bool],
        sort_ascending: Optional[bool],
    ) -> List[Backup]:
        backups = list(self.backups.values())

        if filters:
            for key, values in filters.items():
                if key == "backupIds":
                    backups = [b for b in backups if b.backup_id in values]
                elif key == "sourceBackupIds":
                    backups = [b for b in backups if b.source_backup in values]
                elif key == "clusterIds":
                    backups = [b for b in backups if b.cluster_id in values]
                elif key == "states":
                    backups = [b for b in backups if b.backup_state in values]
                elif key == "neverExpires":
                    never_expires = values[0].lower() == "true"
                    backups = [b for b in backups if b.never_expires == never_expires]

        backups.sort(
            key=lambda x: x.create_timestamp,
            reverse=not sort_ascending if sort_ascending is not None else True,
        )
        return backups

    def put_resource_policy(self, resource_arn: str, policy: str) -> Dict[str, str]:
        self.resource_policies[resource_arn] = policy
        return {"ResourceArn": resource_arn, "Policy": policy}


cloudhsmv2_backends = BackendDict(CloudHSMV2Backend, "cloudhsmv2")



"""LexModelsV2Backend class with methods for supported APIs."""

import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend

from ..utilities.tagging_service import TaggingService


class FakeBot:
    failure_reasons: List[str]

    def __init__(
        self,
        account_id: str,
        region_name: str,
        bot_name: str,
        description: str,
        role_arn: str,
        data_privacy: Optional[Dict[str, Any]],
        idle_session_ttl_in_seconds: int,
        bot_type: str,
        bot_members: Optional[Dict[str, Any]],
    ):
        self.account_id = account_id
        self.region_name = region_name

        self.id = str(uuid.uuid4())
        self.bot_name = bot_name
        self.description = description
        self.role_arn = role_arn
        self.data_privacy = data_privacy
        self.idle_session_ttl_in_seconds = idle_session_ttl_in_seconds
        self.bot_type = bot_type
        self.bot_members = bot_members
        self.status = "CREATING"
        self.creation_date_time = datetime.now().isoformat()
        self.last_updated_date_time = datetime.now().isoformat()
        self.failure_reasons = []

        self.arn = self._generate_arn()

    def _generate_arn(self) -> str:
        return f"arn:aws:lex:{self.region_name}:{self.account_id}:bot/{self.id}"


class FakeBotAlias:
    parent_bot_networks: List[str]
    history_events: List[str]

    def __init__(
        self,
        account_id: str,
        region_name: str,
        bot_alias_name: str,
        description: str,
        bot_version: str,
        bot_alias_locale_settings: Optional[Dict[str, Any]],
        conversation_log_settings: Optional[Dict[str, Any]],
        sentiment_analysis_settings: Optional[Dict[str, Any]],
        bot_id: str,
    ):
        self.account_id = account_id
        self.region_name = region_name

        self.id = str(uuid.uuid4())
        self.bot_alias_name = bot_alias_name
        self.description = description
        self.bot_version = bot_version
        self.bot_alias_locale_settings = bot_alias_locale_settings
        self.conversation_log_settings = conversation_log_settings
        self.sentiment_analysis_settings = sentiment_analysis_settings
        self.status = "CREATING"
        self.bot_id = bot_id
        self.creation_date_time = datetime.now().isoformat()
        self.last_updated_date_time = datetime.now().isoformat()
        self.parent_bot_networks = []
        self.history_events = []

        self.arn = self._generate_arn()

    def _generate_arn(self) -> str:
        return f"arn:aws:lex:{self.region_name}:{self.account_id}:bot-alias/${self.bot_id}/${self.id}"


class FakeResourcePolicy:
    def __init__(self, resource_arn: str, policy: str):
        self.resource_arn = resource_arn
        self.policy = policy
        self.revision_id = str(uuid.uuid4())


class LexModelsV2Backend(BaseBackend):
    """Implementation of LexModelsV2 APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.bots: Dict[str, FakeBot] = {}
        self.bot_aliases: Dict[str, FakeBotAlias] = {}
        self.resource_policies: Dict[str, FakeResourcePolicy] = {}
        self.tagger = TaggingService()

    def create_bot(
        self,
        bot_name: str,
        description: str,
        role_arn: str,
        data_privacy: Optional[Dict[str, Any]],
        idle_session_ttl_in_seconds: int,
        bot_tags: Optional[Dict[str, str]],
        test_bot_alias_tags: Optional[Dict[str, str]],
        bot_type: str,
        bot_members: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        bot = FakeBot(
            account_id=self.account_id,
            region_name=self.region_name,
            bot_name=bot_name,
            description=description,
            role_arn=role_arn,
            data_privacy=data_privacy,
            idle_session_ttl_in_seconds=idle_session_ttl_in_seconds,
            bot_type=bot_type,
            bot_members=bot_members,
        )

        self.bots[bot.id] = bot

        if bot_tags:
            self.tag_resource(bot.arn, bot_tags)

        test_alias = FakeBotAlias(
            account_id=self.account_id,
            region_name=self.region_name,
            bot_alias_name="test",
            description="test",
            bot_version="1",
            bot_alias_locale_settings={},
            conversation_log_settings={},
            sentiment_analysis_settings={},
            bot_id=bot.id,
        )
        self.bot_aliases[test_alias.id] = test_alias

        return {
            "botId": bot.id,
            "botName": bot.bot_name,
            "description": bot.description,
            "roleArn": bot.role_arn,
            "dataPrivacy": bot.data_privacy,
            "idleSessionTTLInSeconds": bot.idle_session_ttl_in_seconds,
            "botStatus": bot.status,
            "creationDateTime": bot.creation_date_time,
            "botTags": self.list_tags_for_resource(bot.arn),
            "testBotAliasTags": test_bot_alias_tags,
            "botType": bot.bot_type,
            "botMembers": bot.bot_members,
        }

    def describe_bot(self, bot_id: str) -> Dict[str, Any]:
        bot = self.bots[bot_id]

        return {
            "botId": bot.id,
            "botName": bot.bot_name,
            "description": bot.description,
            "roleArn": bot.role_arn,
            "dataPrivacy": bot.data_privacy,
            "idleSessionTTLInSeconds": bot.idle_session_ttl_in_seconds,
            "botStatus": bot.status,
            "creationDateTime": bot.creation_date_time,
            "lastUpdatedDateTime": bot.last_updated_date_time,
            "botType": bot.bot_type,
            "botMembers": bot.bot_members,
            "failureReasons": bot.failure_reasons,
        }

    def update_bot(
        self,
        bot_id: str,
        bot_name: str,
        description: str,
        role_arn: str,
        data_privacy: Optional[Dict[str, Any]],
        idle_session_ttl_in_seconds: int,
        bot_type: str,
        bot_members: Optional[Dict[str, Any]],
    ) -> Dict[str, Any]:
        bot = self.bots[bot_id]

        bot.bot_name = bot_name
        bot.description = description
        bot.role_arn = role_arn
        bot.data_privacy = data_privacy
        bot.idle_session_ttl_in_seconds = idle_session_ttl_in_seconds
        bot.bot_type = bot_type
        bot.bot_members = bot_members
        bot.last_updated_date_time = datetime.now().isoformat()

        return {
            "botId": bot.id,
            "botName": bot.bot_name,
            "description": bot.description,
            "roleArn": bot.role_arn,
            "dataPrivacy": bot.data_privacy,
            "idleSessionTTLInSeconds": bot.idle_session_ttl_in_seconds,
            "botStatus": bot.status,
            "creationDateTime": bot.creation_date_time,
            "lastUpdatedDateTime": bot.last_updated_date_time,
            "botType": bot.bot_type,
            "botMembers": bot.bot_members,
        }

    def list_bots(self) -> List[Dict[str, Any]]:
        bot_summaries = [
            {
                "botId": bot.id,
                "botName": bot.bot_name,
                "description": bot.description,
                "botStatus": bot.status,
                "latestBotVersion": 1,
                "lastUpdatedDateTime": bot.last_updated_date_time,
                "botType": bot.bot_type,
            }
            for bot in self.bots.values()
        ]
        return bot_summaries

    def delete_bot(
        self, bot_id: str, skip_resource_in_use_check: bool
    ) -> Tuple[str, str]:
        bot = self.bots.pop(bot_id)
        return bot.id, bot.status

    def create_bot_alias(
        self,
        bot_alias_name: str,
        description: str,
        bot_version: str,
        bot_alias_locale_settings: Optional[Dict[str, Any]],
        conversation_log_settings: Optional[Dict[str, Any]],
        sentiment_analysis_settings: Optional[Dict[str, Any]],
        bot_id: str,
        tags: Optional[Dict[str, str]],
    ) -> Dict[str, Any]:
        bot_alias = FakeBotAlias(
            self.account_id,
            self.region_name,
            bot_alias_name,
            description,
            bot_version,
            bot_alias_locale_settings,
            conversation_log_settings,
            sentiment_analysis_settings,
            bot_id,
        )

        self.bot_aliases[bot_alias.id] = bot_alias

        if tags:
            self.tag_resource(bot_alias.arn, tags)

        return {
            "botAliasId": bot_alias.id,
            "botAliasName": bot_alias.bot_alias_name,
            "description": bot_alias.description,
            "botVersion": bot_alias.bot_version,
            "botAliasLocaleSettings": bot_alias.bot_alias_locale_settings,
            "conversationLogSettings": bot_alias.conversation_log_settings,
            "sentimentAnalysisSettings": bot_alias.sentiment_analysis_settings,
            "botAliasStatus": bot_alias.status,
            "creationDateTime": bot_alias.creation_date_time,
            "botId": bot_alias.bot_id,
            "tags": self.list_tags_for_resource(bot_alias.arn),
        }

    def describe_bot_alias(self, bot_alias_id: str, bot_id: str) -> Dict[str, Any]:
        ba = self.bot_aliases[bot_alias_id]

        return {
            "botAliasId": ba.id,
            "botAliasName": ba.bot_alias_name,
            "description": ba.description,
            "botVersion": ba.bot_version,
            "botAliasLocaleSettings": ba.bot_alias_locale_settings,
            "conversationLogSettings": ba.conversation_log_settings,
            "sentimentAnalysisSettings": ba.sentiment_analysis_settings,
            "botAliasHistoryEvents": ba.history_events,
            "botAliasStatus": ba.status,
            "botId": ba.bot_id,
            "creationDateTime": ba.creation_date_time,
            "lastUpdatedDateTime": ba.last_updated_date_time,
            "parentBotNetworks": ba.parent_bot_networks,
        }

    def update_bot_alias(
        self,
        bot_alias_id: str,
        bot_alias_name: str,
        description: str,
        bot_version: str,
        bot_alias_locale_settings: Optional[Dict[str, Any]],
        conversation_log_settings: Optional[Dict[str, Any]],
        sentiment_analysis_settings: Optional[Dict[str, Any]],
        bot_id: str,
    ) -> Dict[str, Any]:
        ba = self.bot_aliases[bot_alias_id]

        ba.bot_alias_name = bot_alias_name
        ba.description = description
        ba.bot_version = bot_version
        ba.bot_alias_locale_settings = bot_alias_locale_settings
        ba.conversation_log_settings = conversation_log_settings
        ba.sentiment_analysis_settings = sentiment_analysis_settings
        ba.bot_id = bot_id
        ba.last_updated_date_time = datetime.now().isoformat()

        return {
            "botAliasId": ba.id,
            "botAliasName": ba.bot_alias_name,
            "description": ba.description,
            "botVersion": ba.bot_version,
            "botAliasLocaleSettings": ba.bot_alias_locale_settings,
            "conversationLogSettings": ba.conversation_log_settings,
            "sentimentAnalysisSettings": ba.sentiment_analysis_settings,
            "botAliasStatus": ba.status,
            "botId": ba.bot_id,
            "creationDateTime": ba.creation_date_time,
            "lastUpdatedDateTime": ba.last_updated_date_time,
        }

    def list_bot_aliases(
        self, bot_id: str, max_results: int
    ) -> Tuple[List[Dict[str, Any]], Optional[str]]:
        bot_alias_summaries = [
            {
                "botAliasId": ba.id,
                "botAliasName": ba.bot_alias_name,
                "description": ba.description,
                "botVersion": ba.bot_version,
                "botAliasStatus": ba.status,
                "creationDateTime": ba.creation_date_time,
                "lastUpdatedDateTime": ba.last_updated_date_time,
            }
            for ba in self.bot_aliases.values()
        ]

        return bot_alias_summaries, bot_id

    def delete_bot_alias(
        self, bot_alias_id: str, bot_id: str, skip_resource_in_use_check: bool
    ) -> Tuple[str, str, str]:
        ba = self.bot_aliases.pop(bot_alias_id)
        return ba.id, ba.bot_id, ba.status

    def create_resource_policy(self, resource_arn: str, policy: str) -> Tuple[str, str]:
        rp = FakeResourcePolicy(resource_arn, policy)
        self.resource_policies[rp.resource_arn] = rp
        return rp.resource_arn, rp.revision_id

    def describe_resource_policy(self, resource_arn: str) -> Tuple[str, str, str]:
        rp = self.resource_policies[resource_arn]
        return rp.resource_arn, rp.policy, rp.revision_id

    def update_resource_policy(
        self, resource_arn: str, policy: str, expected_revision_id: str
    ) -> Tuple[str, str]:
        rp = self.resource_policies[resource_arn]
        if expected_revision_id != rp.revision_id:
            raise Exception("Revision ID mismatch")
        rp.policy = policy
        rp.revision_id = str(uuid.uuid4())
        return rp.resource_arn, rp.revision_id

    def delete_resource_policy(
        self, resource_arn: str, expected_revision_id: str
    ) -> Tuple[str, str]:
        rp = self.resource_policies[resource_arn]
        if expected_revision_id != rp.revision_id:
            raise Exception("Revision ID mismatch")
        rp = self.resource_policies.pop(resource_arn)
        return rp.resource_arn, rp.revision_id

    def tag_resource(self, resource_arn: str, tags: Dict[str, str]) -> None:
        tags_list = [{"Key": k, "Value": v} for k, v in tags.items()]
        self.tagger.tag_resource(resource_arn, tags_list)

    def untag_resource(self, resource_arn: str, tag_keys: List[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> Dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)


lexv2models_backends = BackendDict(LexModelsV2Backend, "lexv2-models")



from uuid import uuid4

import boto3
import pytest

from . import lakeformation_aws_verified


@pytest.mark.aws_verified
@lakeformation_aws_verified
def test_add_unknown_lf_tags(
    bucket_name=None,  # pylint: disable=unused-argument
    db_name=None,
    table_name=None,  # pylint: disable=unused-argument
    column_name=None,  # pylint: disable=unused-argument
):
    client = boto3.client("lakeformation", region_name="eu-west-2")
    sts = boto3.client("sts", "eu-west-2")
    account_id = sts.get_caller_identity()["Account"]

    failures = client.add_lf_tags_to_resource(
        Resource={"Database": {"Name": db_name}},
        LFTags=[{"TagKey": "unknown-tag", "TagValues": ["value"]}],
    )["Failures"]
    assert len(failures) == 1
    assert failures[0]["LFTag"] == {
        "CatalogId": account_id,
        "TagKey": "unknown-tag",
        "TagValues": ["value"],
    }
    assert failures[0]["Error"] == {
        "ErrorCode": "EntityNotFoundException",
        "ErrorMessage": "Tag or tag value does not exist.",
    }


@pytest.mark.aws_verified
@lakeformation_aws_verified
def test_tag_lakeformation_database(
    bucket_name=None,  # pylint: disable=unused-argument
    db_name=None,
    table_name=None,  # pylint: disable=unused-argument
    column_name=None,  # pylint: disable=unused-argument
):
    client = boto3.client("lakeformation", region_name="eu-west-2")
    sts = boto3.client("sts", "eu-west-2")
    account_id = sts.get_caller_identity()["Account"]

    tag_name = str(uuid4())[0:6]
    client.create_lf_tag(TagKey=tag_name, TagValues=["value1"])
    resp = client.add_lf_tags_to_resource(
        Resource={"Database": {"Name": db_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )
    assert resp["Failures"] == []

    tags = client.get_resource_lf_tags(Resource={"Database": {"Name": db_name}})[
        "LFTagOnDatabase"
    ]
    assert tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    client.update_lf_tag(TagKey=tag_name, TagValuesToAdd=["value2"])

    all_tags = client.list_lf_tags()["LFTags"]
    our_tag = next(tag for tag in all_tags if tag["TagKey"] == tag_name)
    assert set(our_tag["TagValues"]) == {"value1", "value2"}

    # The value for this particular resource has not been updated
    db_tags = client.get_resource_lf_tags(Resource={"Database": {"Name": db_name}})[
        "LFTagOnDatabase"
    ]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    # Update the existing tags for this resource
    client.add_lf_tags_to_resource(
        Resource={"Database": {"Name": db_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    db_tags = client.get_resource_lf_tags(Resource={"Database": {"Name": db_name}})[
        "LFTagOnDatabase"
    ]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value2"]}
    ]

    # Try remove and re-add
    client.remove_lf_tags_from_resource(
        Resource={"Database": {"Name": db_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    assert "LFTagOnDatabase" not in client.get_resource_lf_tags(
        Resource={"Database": {"Name": db_name}}
    )

    client.add_lf_tags_to_resource(
        Resource={"Database": {"Name": db_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )

    db_tags = client.get_resource_lf_tags(Resource={"Database": {"Name": db_name}})[
        "LFTagOnDatabase"
    ]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    # Deleting the tag automatically deletes it from any resource
    client.delete_lf_tag(TagKey=tag_name)

    assert "LFTagOnDatabase" not in client.get_resource_lf_tags(
        Resource={"Database": {"Name": db_name}}
    )


@pytest.mark.aws_verified
@lakeformation_aws_verified
def test_tag_lakeformation_table(
    bucket_name=None,  # pylint: disable=unused-argument
    db_name=None,
    table_name=None,
    column_name=None,  # pylint: disable=unused-argument
):
    client = boto3.client("lakeformation", region_name="eu-west-2")
    sts = boto3.client("sts", "eu-west-2")
    account_id = sts.get_caller_identity()["Account"]

    tag_name = str(uuid4())[0:6]
    client.create_lf_tag(TagKey=tag_name, TagValues=["value1"])
    resp = client.add_lf_tags_to_resource(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )
    assert resp["Failures"] == []

    tags = client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
    )["LFTagsOnTable"]
    assert tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    client.update_lf_tag(TagKey=tag_name, TagValuesToAdd=["value2"])

    all_tags = client.list_lf_tags()["LFTags"]
    our_tag = next(tag for tag in all_tags if tag["TagKey"] == tag_name)
    assert set(our_tag["TagValues"]) == {"value1", "value2"}

    # The value for this particular resource has not been updated
    db_tags = client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
    )["LFTagsOnTable"]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    # Update the existing tags for this resource
    client.add_lf_tags_to_resource(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    db_tags = client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}}
    )["LFTagsOnTable"]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value2"]}
    ]

    # Try remove and re-add
    client.remove_lf_tags_from_resource(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    assert "LFTagsOnTable" not in client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}}
    )

    client.add_lf_tags_to_resource(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}},
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )

    db_tags = client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}}
    )["LFTagsOnTable"]
    assert db_tags == [
        {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
    ]

    # Deleting the tag automatically deletes it from any resource
    client.delete_lf_tag(TagKey=tag_name)

    assert "LFTagsOnTable" not in client.get_resource_lf_tags(
        Resource={"Table": {"DatabaseName": db_name, "Name": table_name}}
    )


@pytest.mark.aws_verified
@lakeformation_aws_verified
def test_tag_lakeformation_columns(
    bucket_name=None,  # pylint: disable=unused-argument
    db_name=None,
    table_name=None,
    column_name=None,
):
    client = boto3.client("lakeformation", region_name="eu-west-2")
    sts = boto3.client("sts", "eu-west-2")
    account_id = sts.get_caller_identity()["Account"]

    tag_name = str(uuid4())[0:6]
    client.create_lf_tag(TagKey=tag_name, TagValues=["value1"])
    resp = client.add_lf_tags_to_resource(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )
    assert resp["Failures"] == []

    tags = client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
    )["LFTagsOnColumns"]
    assert tags == [
        {
            "Name": column_name,
            "LFTags": [
                {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
            ],
        }
    ]

    client.update_lf_tag(TagKey=tag_name, TagValuesToAdd=["value2"])

    all_tags = client.list_lf_tags()["LFTags"]
    our_tag = next(tag for tag in all_tags if tag["TagKey"] == tag_name)
    assert set(our_tag["TagValues"]) == {"value1", "value2"}

    # The value for this particular resource has not been updated
    tags = client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
    )["LFTagsOnColumns"]
    assert tags == [
        {
            "Name": column_name,
            "LFTags": [
                {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
            ],
        }
    ]

    # Update the existing tags for this resource
    client.add_lf_tags_to_resource(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    tags = client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        }
    )["LFTagsOnColumns"]
    assert tags == [
        {
            "Name": column_name,
            "LFTags": [
                {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value2"]}
            ],
        }
    ]

    # Try remove and re-add
    client.remove_lf_tags_from_resource(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
        LFTags=[{"TagKey": tag_name, "TagValues": ["value2"]}],
    )

    assert "LFTagsOnColumns" not in client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        }
    )

    client.add_lf_tags_to_resource(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        },
        LFTags=[{"TagKey": tag_name, "TagValues": ["value1"]}],
    )

    tags = client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        }
    )["LFTagsOnColumns"]
    assert tags == [
        {
            "Name": column_name,
            "LFTags": [
                {"CatalogId": account_id, "TagKey": tag_name, "TagValues": ["value1"]}
            ],
        }
    ]

    # Deleting the tag automatically deletes it from any resource
    client.delete_lf_tag(TagKey=tag_name)

    assert "LFTagsOnColumns" not in client.get_resource_lf_tags(
        Resource={
            "TableWithColumns": {
                "DatabaseName": db_name,
                "Name": table_name,
                "ColumnNames": [column_name],
            }
        }
    )


@pytest.mark.aws_verified
@lakeformation_aws_verified
def test_lf_tags(bucket_name=None, db_name=None, table_name=None, column_name=None):  # pylint: disable=unused-argument
    client = boto3.client("lakeformation", region_name="eu-west-2")
    sts = boto3.client("sts", "eu-west-2")
    account_id = sts.get_caller_identity()["Account"]

    client.create_lf_tag(TagKey="tag1", TagValues=["1a", "1b", "1c"])
    client.create_lf_tag(TagKey="tag2", TagValues=["2a", "2b"])
    client.create_lf_tag(TagKey="tag3", TagValues=["3a", "3b"])

    resp = client.get_lf_tag(TagKey="tag1")
    assert resp["CatalogId"] == account_id
    assert resp["TagKey"] == "tag1"
    assert resp["TagValues"] == ["1a", "1b", "1c"]

    client.update_lf_tag(TagKey="tag1", TagValuesToDelete=["1a", "1c"])

    tags = client.list_lf_tags()["LFTags"]
    assert set([x["CatalogId"] for x in tags]) == {account_id}
    tag_keys = [x["TagKey"] for x in tags]
    assert "tag1" in tag_keys
    assert "tag2" in tag_keys
    assert "tag3" in tag_keys

    assert [x for x in tags if x["TagKey"] == "tag1"][0]["TagValues"] == ["1b"]
    assert set([x for x in tags if x["TagKey"] == "tag2"][0]["TagValues"]) == {
        "2a",
        "2b",
    }
    assert set([x for x in tags if x["TagKey"] == "tag3"][0]["TagValues"]) == {
        "3a",
        "3b",
    }

    client.delete_lf_tag(TagKey="tag2")

    tags = client.list_lf_tags()["LFTags"]
    tag_keys = [x["TagKey"] for x in tags]
    assert "tag1" in tag_keys
    assert "tag3" in tag_keys
    assert "tag2" not in tag_keys

    client.delete_lf_tag(TagKey="tag1")
    client.delete_lf_tag(TagKey="tag3")



import json

import boto3

from moto import mock_aws


@mock_aws
def test_search_things_include_named_shadow():
    iot_client = boto3.client("iot", region_name="ap-northeast-1")
    iotdata_client = boto3.client("iot-data", region_name="ap-northeast-1")
    raw_payload = b'{"state": {"desired": {"led": "on"}, "reported": {"led": "off"}}}'

    thing_name = "test-thing-name"
    iot_client.create_thing(thingName=thing_name)
    iotdata_client.update_thing_shadow(
        thingName=thing_name, shadowName="test_shadow", payload=raw_payload
    )

    resp = iot_client.search_index(queryString=f"thingName:{thing_name}")

    assert len(resp["things"]) == 1
    shadow = json.loads(resp["things"][0]["shadow"])

    assert shadow["name"]["test_shadow"]["desired"] == {"led": "on"}
    assert shadow["name"]["test_shadow"]["reported"] == {"led": "off"}
    assert shadow["name"]["test_shadow"]["hasDelta"]



import json
from functools import wraps
from time import sleep
from typing import TYPE_CHECKING, Callable, TypeVar
from uuid import uuid4

import boto3
import requests

from moto import mock_aws, settings
from tests import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from tests import allow_aws_request
from tests.test_stepfunctions.parser.templates.templates import load_template

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")

T = TypeVar("T")


def aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.
    """

    @wraps(func)
    def pagination_wrapper():
        if allow_aws_request():
            return func()
        else:
            with mock_aws():
                requests.post(
                    f"http://{base_url}/moto-api/config",
                    json={"stepfunctions": {"execute_state_machine": True}},
                )
                resp = func()
                requests.post(
                    f"http://{base_url}/moto-api/config",
                    json={"stepfunctions": {"execute_state_machine": False}},
                )
                return resp

    return pagination_wrapper


def verify_execution_result(
    _verify_result, expected_status, tmpl_name, exec_input=None, sleep_time=0
):
    iam = boto3.client("iam", region_name="us-east-1")
    role_name = f"sfn_role_{str(uuid4())[0:6]}"
    sfn_role = iam.create_role(
        RoleName=role_name,
        AssumeRolePolicyDocument=json.dumps(sfn_role_policy),
        Path="/",
    )["Role"]["Arn"]
    iam.put_role_policy(
        PolicyDocument=json.dumps(sfn_allow_dynamodb),
        PolicyName="allowLambdaInvoke",
        RoleName=role_name,
    )
    sleep(sleep_time)

    client = boto3.client("stepfunctions", region_name="us-east-1")
    execution_arn, state_machine_arn = _start_execution(
        client, load_template(tmpl_name), exec_input, sfn_role
    )
    for _ in range(30):
        execution = client.describe_execution(executionArn=execution_arn)
        if expected_status is None or execution["status"] == expected_status:
            result = _verify_result(client, execution, execution_arn)
            if result is not False:
                client.delete_state_machine(stateMachineArn=state_machine_arn)
                iam.delete_role_policy(
                    RoleName=role_name, PolicyName="allowLambdaInvoke"
                )
                iam.delete_role(RoleName=role_name)
                break
        sleep(10 if allow_aws_request() else 0.1)
    else:
        client.delete_state_machine(stateMachineArn=state_machine_arn)
        iam.delete_role_policy(RoleName=role_name, PolicyName="allowLambdaInvoke")
        iam.delete_role(RoleName=role_name)
        assert False, "Should have failed already"


def _start_execution(client, definition, exec_input, sfn_role):
    name = f"sfn_name_{str(uuid4())[0:6]}"
    #
    response = client.create_state_machine(
        name=name, definition=json.dumps(definition), roleArn=sfn_role
    )
    state_machine_arn = response["stateMachineArn"]
    if exec_input:
        execution = client.start_execution(
            name="exec1", stateMachineArn=state_machine_arn, input=exec_input or "{}"
        )
    else:
        execution = client.start_execution(
            name="exec1", stateMachineArn=state_machine_arn
        )
    execution_arn = execution["executionArn"]
    return execution_arn, state_machine_arn


def _get_default_role():
    return "arn:aws:iam::" + ACCOUNT_ID + ":role/unknown_sf_role"


base_url = "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"


def enable_sfn_parser(func: "Callable[P, T]") -> "Callable[P, T]":
    @wraps(func)
    def pagination_wrapper(*args: "P.args", **kwargs: "P.kwargs") -> T:  # type: ignore
        requests.post(
            f"http://{base_url}/moto-api/config",
            json={"stepfunctions": {"execute_state_machine": True}},
        )
        try:
            res = func(*args, **kwargs)
        finally:
            requests.post(
                f"http://{base_url}/moto-api/config",
                json={"stepfunctions": {"execute_state_machine": False}},
            )
        return res

    return pagination_wrapper


sfn_role_policy = {
    "Version": "2012-10-17",
    "Statement": [
        {
            "Effect": "Allow",
            "Principal": {"Service": "states.amazonaws.com"},
            "Action": "sts:AssumeRole",
        }
    ],
}

sfn_allow_lambda_invoke = {
    "Version": "2012-10-17",
    "Statement": [
        {"Effect": "Allow", "Action": ["lambda:InvokeFunction"], "Resource": ["*"]}
    ],
}

sfn_allow_dynamodb = {
    "Version": "2012-10-17",
    "Statement": [{"Effect": "Allow", "Action": ["dynamodb:*"], "Resource": ["*"]}],
}



"""Handles incoming workspacesweb requests, invokes methods, returns responses."""

import json
from typing import Any
from urllib.parse import unquote

from moto.core.responses import TYPE_RESPONSE, BaseResponse

from .models import WorkSpacesWebBackend, workspacesweb_backends


class WorkSpacesWebResponse(BaseResponse):
    """Handler for WorkSpacesWeb requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="workspaces-web")

    @property
    def workspacesweb_backend(self) -> WorkSpacesWebBackend:
        """Return backend instance specific for this region."""
        return workspacesweb_backends[self.current_account][self.region]

    @staticmethod
    def network_settings(request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[misc]
        handler = WorkSpacesWebResponse()
        handler.setup_class(request, full_url, headers)
        if request.method == "GET":
            return handler.get_network_settings()
        else:
            return handler.delete_network_settings()

    @staticmethod
    def browser_settings(request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[misc]
        handler = WorkSpacesWebResponse()
        handler.setup_class(request, full_url, headers)
        if request.method == "GET":
            return handler.get_browser_settings()
        else:
            return handler.delete_browser_settings()

    @staticmethod
    def user_settings(request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[misc]
        handler = WorkSpacesWebResponse()
        handler.setup_class(request, full_url, headers)
        if request.method == "GET":
            return handler.get_user_settings()
        else:
            return handler.delete_user_settings()

    @staticmethod
    def user_access_logging_settings(  # type: ignore[misc]
        request: Any, full_url: str, headers: Any
    ) -> TYPE_RESPONSE:
        handler = WorkSpacesWebResponse()
        handler.setup_class(request, full_url, headers)
        if request.method == "GET":
            return handler.get_user_access_logging_settings()
        else:
            return handler.delete_user_access_logging_settings()

    @staticmethod
    def portal(request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore[misc]
        handler = WorkSpacesWebResponse()
        handler.setup_class(request, full_url, headers)
        if request.method == "GET":
            return handler.get_portal()
        else:
            return handler.delete_portal()

    def create_browser_settings(self) -> str:
        additional_encryption_context = self._get_param("additionalEncryptionContext")
        browser_policy = self._get_param("browserPolicy")
        client_token = self._get_param("clientToken")
        customer_managed_key = self._get_param("customerManagedKey")
        tags = self._get_param("tags")
        browser_settings_arn = self.workspacesweb_backend.create_browser_settings(
            additional_encryption_context=additional_encryption_context,
            browser_policy=browser_policy,
            client_token=client_token,
            customer_managed_key=customer_managed_key,
            tags=tags,
        )
        return json.dumps(dict(browserSettingsArn=browser_settings_arn))

    def create_network_settings(self) -> str:
        security_group_ids = self._get_param("securityGroupIds")
        subnet_ids = self._get_param("subnetIds")
        tags = self._get_param("tags")
        vpc_id = self._get_param("vpcId")
        network_settings_arn = self.workspacesweb_backend.create_network_settings(
            security_group_ids=security_group_ids,
            subnet_ids=subnet_ids,
            tags=tags,
            vpc_id=vpc_id,
        )
        return json.dumps(dict(networkSettingsArn=network_settings_arn))

    def get_network_settings(self) -> TYPE_RESPONSE:
        network_settings_arn = unquote(
            self.parsed_url.path.split("/networkSettings/")[-1]
        )
        network_settings = self.workspacesweb_backend.get_network_settings(
            network_settings_arn=network_settings_arn,
        )
        return 200, {}, json.dumps(dict(networkSettings=network_settings))

    def create_portal(self) -> str:
        additional_encryption_context = self._get_param("additionalEncryptionContext")
        authentication_type = self._get_param("authenticationType")
        client_token = self._get_param("clientToken")
        customer_managed_key = self._get_param("customerManagedKey")
        display_name = self._get_param("displayName")
        instance_type = self._get_param("instanceType")
        max_concurrent_sessions = self._get_param("maxConcurrentSessions")
        tags = self._get_param("tags")
        portal_arn, portal_endpoint = self.workspacesweb_backend.create_portal(
            additional_encryption_context=additional_encryption_context,
            authentication_type=authentication_type,
            client_token=client_token,
            customer_managed_key=customer_managed_key,
            display_name=display_name,
            instance_type=instance_type,
            max_concurrent_sessions=max_concurrent_sessions,
            tags=tags,
        )
        return json.dumps(dict(portalArn=portal_arn, portalEndpoint=portal_endpoint))

    def list_browser_settings(self) -> str:
        browser_settings = self.workspacesweb_backend.list_browser_settings()
        return json.dumps(dict(browserSettings=browser_settings))

    def list_network_settings(self) -> str:
        network_settings = self.workspacesweb_backend.list_network_settings()
        return json.dumps(dict(networkSettings=network_settings))

    def list_portals(self) -> str:
        portals = self.workspacesweb_backend.list_portals()
        return json.dumps(dict(portals=portals))

    def get_browser_settings(self) -> TYPE_RESPONSE:
        browser_settings_arn = unquote(
            self.parsed_url.path.split("/browserSettings/")[-1]
        )
        browser_settings = self.workspacesweb_backend.get_browser_settings(
            browser_settings_arn=browser_settings_arn,
        )
        return 200, {}, json.dumps(dict(browserSettings=browser_settings))

    def delete_browser_settings(self) -> TYPE_RESPONSE:
        browser_settings_arn = unquote(
            self.parsed_url.path.split("/browserSettings/")[-1]
        )
        self.workspacesweb_backend.delete_browser_settings(
            browser_settings_arn=browser_settings_arn
        )
        return 200, {}, "{}"

    def delete_network_settings(self) -> TYPE_RESPONSE:
        network_settings_arn = unquote(
            self.parsed_url.path.split("/networkSettings/")[-1]
        )
        self.workspacesweb_backend.delete_network_settings(
            network_settings_arn=network_settings_arn,
        )
        return 200, {}, "{}"

    def get_portal(self) -> TYPE_RESPONSE:
        portal_arn = unquote(self.parsed_url.path.split("/portals/")[-1])
        portal = self.workspacesweb_backend.get_portal(portal_arn=portal_arn)
        return 200, {}, json.dumps(dict(portal=portal))

    def delete_portal(self) -> TYPE_RESPONSE:
        portal_arn = unquote(self.parsed_url.path.split("/portals/")[-1])
        self.workspacesweb_backend.delete_portal(portal_arn=portal_arn)
        return 200, {}, "{}"

    def associate_browser_settings(self) -> str:
        browser_settings_arn = unquote(self._get_param("browserSettingsArn"))
        portal_arn = unquote(
            self.parsed_url.path.split("/portals/")[-1].split("/browserSettings")[0]
        )
        browser_settings_arn, portal_arn = (
            self.workspacesweb_backend.associate_browser_settings(
                browser_settings_arn=browser_settings_arn,
                portal_arn=portal_arn,
            )
        )
        return json.dumps(
            dict(browserSettingsArn=browser_settings_arn, portalArn=portal_arn)
        )

    def associate_network_settings(self) -> str:
        network_settings_arn = unquote(self._get_param("networkSettingsArn"))
        portal_arn = unquote(
            self.parsed_url.path.split("/portals/")[-1].split("/networkSettings")[0]
        )
        network_settings_arn, portal_arn = (
            self.workspacesweb_backend.associate_network_settings(
                network_settings_arn=network_settings_arn,
                portal_arn=portal_arn,
            )
        )
        return json.dumps(
            dict(networkSettingsArn=network_settings_arn, portalArn=portal_arn)
        )

    def create_user_settings(self) -> str:
        additional_encryption_context = self._get_param("additionalEncryptionContext")
        client_token = self._get_param("clientToken")
        cookie_synchronization_configuration = self._get_param(
            "cookieSynchronizationConfiguration"
        )
        copy_allowed = self._get_param("copyAllowed")
        customer_managed_key = self._get_param("customerManagedKey")
        deep_link_allowed = self._get_param("deepLinkAllowed")
        disconnect_timeout_in_minutes = self._get_param("disconnectTimeoutInMinutes")
        download_allowed = self._get_param("downloadAllowed")
        idle_disconnect_timeout_in_minutes = self._get_param(
            "idleDisconnectTimeoutInMinutes"
        )
        paste_allowed = self._get_param("pasteAllowed")
        print_allowed = self._get_param("printAllowed")
        tags = self._get_param("tags")
        upload_allowed = self._get_param("uploadAllowed")
        user_settings_arn = self.workspacesweb_backend.create_user_settings(
            additional_encryption_context=additional_encryption_context,
            client_token=client_token,
            cookie_synchronization_configuration=cookie_synchronization_configuration,
            copy_allowed=copy_allowed,
            customer_managed_key=customer_managed_key,
            deep_link_allowed=deep_link_allowed,
            disconnect_timeout_in_minutes=disconnect_timeout_in_minutes,
            download_allowed=download_allowed,
            idle_disconnect_timeout_in_minutes=idle_disconnect_timeout_in_minutes,
            paste_allowed=paste_allowed,
            print_allowed=print_allowed,
            tags=tags,
            upload_allowed=upload_allowed,
        )
        return json.dumps(dict(userSettingsArn=user_settings_arn))

    def get_user_settings(self) -> TYPE_RESPONSE:
        user_settings_arn = unquote(self.parsed_url.path.split("/userSettings/")[-1])
        user_settings = self.workspacesweb_backend.get_user_settings(
            user_settings_arn=user_settings_arn,
        )
        return 200, {}, json.dumps(dict(userSettings=user_settings))

    def delete_user_settings(self) -> TYPE_RESPONSE:
        user_settings_arn = unquote(self.parsed_url.path.split("/userSettings/")[-1])
        self.workspacesweb_backend.delete_user_settings(
            user_settings_arn=user_settings_arn,
        )
        return 200, {}, "{}"

    def create_user_access_logging_settings(self) -> str:
        params = self._get_params()
        params = json.loads(list(params.keys())[0])
        client_token = params.get("clientToken")
        kinesis_stream_arn = params.get("kinesisStreamArn")
        tags = params.get("tags")
        user_access_logging_settings_arn = (
            self.workspacesweb_backend.create_user_access_logging_settings(
                client_token=client_token,
                kinesis_stream_arn=kinesis_stream_arn,
                tags=tags,
            )
        )
        return json.dumps(
            dict(userAccessLoggingSettingsArn=user_access_logging_settings_arn)
        )

    def get_user_access_logging_settings(self) -> TYPE_RESPONSE:
        user_access_logging_settings_arn = unquote(
            self.parsed_url.path.split("/userAccessLoggingSettings/")[-1]
        )
        user_access_logging_settings = (
            self.workspacesweb_backend.get_user_access_logging_settings(
                user_access_logging_settings_arn=user_access_logging_settings_arn,
            )
        )
        return (
            200,
            {},
            json.dumps(dict(userAccessLoggingSettings=user_access_logging_settings)),
        )

    def delete_user_access_logging_settings(self) -> TYPE_RESPONSE:
        user_access_logging_settings_arn = unquote(
            self.parsed_url.path.split("/userAccessLoggingSettings/")[-1]
        )
        self.workspacesweb_backend.delete_user_access_logging_settings(
            user_access_logging_settings_arn=user_access_logging_settings_arn,
        )
        return 200, {}, "{}"

    def associate_user_settings(self) -> str:
        user_settings_arn = unquote(self._get_param("userSettingsArn"))
        portal_arn = unquote(
            self.parsed_url.path.split("/portals/")[-1].split("/userSettings")[0]
        )
        user_settings_arn, portal_arn = (
            self.workspacesweb_backend.associate_user_settings(
                user_settings_arn=user_settings_arn,
                portal_arn=portal_arn,
            )
        )
        return json.dumps(dict(userSettingsArn=user_settings_arn, portalArn=portal_arn))

    def associate_user_access_logging_settings(self) -> str:
        user_access_logging_settings_arn = unquote(
            self._get_param("userAccessLoggingSettingsArn")
        )
        portal_arn = unquote(
            self.parsed_url.path.split("/portals/")[-1].split(
                "/userAccessLoggingSettings"
            )[0]
        )
        user_access_logging_settings_arn, portal_arn = (
            self.workspacesweb_backend.associate_user_access_logging_settings(
                user_access_logging_settings_arn=user_access_logging_settings_arn,
                portal_arn=portal_arn,
            )
        )
        return json.dumps(
            dict(
                userAccessLoggingSettingsArn=user_access_logging_settings_arn,
                portalArn=portal_arn,
            )
        )

    def list_user_settings(self) -> str:
        user_settings = self.workspacesweb_backend.list_user_settings()
        return json.dumps(dict(userSettings=user_settings))

    def list_user_access_logging_settings(self) -> str:
        user_access_logging_settings = (
            self.workspacesweb_backend.list_user_access_logging_settings()
        )
        return json.dumps(dict(userAccessLoggingSettings=user_access_logging_settings))

    def tag_resource(self) -> str:
        client_token = self._get_param("clientToken")
        resource_arn = unquote(self._get_param("resourceArn"))
        tags = self._get_param("tags")
        self.workspacesweb_backend.tag_resource(
            client_token=client_token,
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps(dict())

    def untag_resource(self) -> str:
        tagKeys = self.__dict__["data"]["tagKeys"]
        resource_arn = unquote(self._get_param("resourceArn"))
        self.workspacesweb_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tagKeys,
        )
        return json.dumps(dict())

    def list_tags_for_resource(self) -> str:
        resource_arn = unquote(self.parsed_url.path.split("/tags/")[-1])
        tags = self.workspacesweb_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps(dict(tags=tags))



"""dsql base URL and path."""

from .responses import AuroraDSQLResponse

url_bases = [
    r"https?://dsql\.(.+)\.api\.aws",
]

url_paths = {
    "{0}/cluster$": AuroraDSQLResponse.dispatch,
    "{0}/cluster/(?P<identifier>[^/]+)$": AuroraDSQLResponse.dispatch,
}



import boto3
import pytest
from botocore.exceptions import ClientError
from freezegun import freeze_time

from moto import mock_aws


@freeze_time("2015-01-01")
@mock_aws
def test_create_layer_response():
    client = boto3.client("opsworks", region_name="us-east-1")
    stack_id = client.create_stack(
        Name="test_stack_1",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn",
    )["StackId"]

    response = client.create_layer(
        StackId=stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName",
    )

    assert "LayerId" in response

    second_stack_id = client.create_stack(
        Name="test_stack_2",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn",
    )["StackId"]

    response = client.create_layer(
        StackId=second_stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName",
    )

    assert "LayerId" in response

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.create_layer(
            StackId=stack_id, Type="custom", Name="TestLayer", Shortname="_"
        )
    assert (
        r'already a layer named "TestLayer"' in exc.value.response["Error"]["Message"]
    )

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.create_layer(
            StackId=stack_id, Type="custom", Name="_", Shortname="TestLayerShortName"
        )
    assert (
        r'already a layer with shortname "TestLayerShortName"'
    ) in exc.value.response["Error"]["Message"]

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.create_layer(
            StackId="nothere", Type="custom", Name="TestLayer", Shortname="_"
        )
    assert exc.value.response["Error"]["Message"] == "nothere"


@freeze_time("2015-01-01")
@mock_aws
def test_describe_layers():
    client = boto3.client("opsworks", region_name="us-east-1")
    stack_id = client.create_stack(
        Name="test_stack_1",
        Region="us-east-1",
        ServiceRoleArn="service_arn",
        DefaultInstanceProfileArn="profile_arn",
    )["StackId"]
    layer_id = client.create_layer(
        StackId=stack_id,
        Type="custom",
        Name="TestLayer",
        Shortname="TestLayerShortName",
    )["LayerId"]

    rv1 = client.describe_layers(StackId=stack_id)
    rv2 = client.describe_layers(LayerIds=[layer_id])
    assert rv1["Layers"] == rv2["Layers"]

    assert rv1["Layers"][0]["Name"] == "TestLayer"

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.describe_layers(StackId=stack_id, LayerIds=[layer_id])
    assert exc.value.response["Error"]["Message"] == (
        "Please provide one or more layer IDs or a stack ID"
    )

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.describe_layers(StackId="nothere")
    assert exc.value.response["Error"]["Message"] == (
        "Unable to find stack with ID nothere"
    )

    # ClientError
    with pytest.raises(ClientError) as exc:
        client.describe_layers(LayerIds=["nothere"])
    assert exc.value.response["Error"]["Message"] == "nothere"



from .responses import ManagedBlockchainResponse

url_bases = [r"https?://managedblockchain\.(.+)\.amazonaws.com"]

url_paths = {
    "{0}/networks$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/proposals$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/proposals/(?P<proposalid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/proposals/(?P<proposalid>[^/.]+)/votes$": ManagedBlockchainResponse.dispatch,
    "{0}/invitations$": ManagedBlockchainResponse.dispatch,
    "{0}/invitations/(?P<invitationid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/members$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/members/(?P<memberid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/members/(?P<memberid>[^/.]+)/nodes$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/members/(?P<memberid>[^/.]+)/nodes?(?P<querys>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/members/(?P<memberid>[^/.]+)/nodes/(?P<nodeid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
    # >= botocore 1.19.41 (API change - memberId is now part of query-string or body)
    "{0}/networks/(?P<networkid>[^/.]+)/nodes$": ManagedBlockchainResponse.dispatch,
    "{0}/networks/(?P<networkid>[^/.]+)/nodes/(?P<nodeid>[^/.]+)$": ManagedBlockchainResponse.dispatch,
}



"""Exceptions raised by the servicequotas service."""

from moto.core.exceptions import JsonRESTError


class NoSuchResource(JsonRESTError):
    def __init__(self) -> None:
        super().__init__(
            "NoSuchResourceException",
            "This service is not available in the current Region. Choose a different Region or a different service.",
        )



"""Handles incoming opensearch requests, invokes methods, returns responses."""

import json
import re
from typing import Any

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse
from moto.es.exceptions import InvalidDomainName

from .models import OpenSearchServiceBackend, opensearch_backends


class OpenSearchServiceResponse(BaseResponse):
    """Handler for OpenSearchService requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="opensearch")

    @property
    def opensearch_backend(self) -> OpenSearchServiceBackend:
        """Return backend instance specific for this region."""
        return opensearch_backends[self.current_account][self.region]

    @classmethod
    def list_domains(cls, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore
        response = cls()
        response.setup_class(request, full_url, headers)
        if request.method == "GET":
            return 200, {}, response.list_domain_names()
        if request.method == "POST":
            return 200, {}, response.describe_domains()

    @classmethod
    def domains(cls, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore
        response = cls()
        response.setup_class(request, full_url, headers)
        if request.method == "POST":
            return 200, {}, response.create_domain()

    @classmethod
    def domain(cls, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore
        response = cls()
        response.setup_class(request, full_url, headers)
        if request.method == "DELETE":
            return 200, {}, response.delete_domain()
        if request.method == "GET":
            return 200, {}, response.describe_domain()

    @classmethod
    def tags(cls, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore
        response = cls()
        response.setup_class(request, full_url, headers)
        if request.method == "GET":
            return 200, {}, response.list_tags()
        if request.method == "POST":
            return 200, {}, response.add_tags()

    @classmethod
    def tag_removal(cls, request: Any, full_url: str, headers: Any) -> TYPE_RESPONSE:  # type: ignore
        response = cls()
        response.setup_class(request, full_url, headers)
        if request.method == "POST":
            return 200, {}, response.remove_tags()

    def create_domain(self) -> str:
        domain_name = self._get_param("DomainName")
        if not re.match(r"^[a-z][a-z0-9\-]+$", domain_name):
            raise InvalidDomainName(domain_name)
        engine_version = self._get_param("EngineVersion")
        cluster_config = self._get_param("ClusterConfig")
        ebs_options = self._get_param("EBSOptions")
        access_policies = self._get_param("AccessPolicies")
        snapshot_options = self._get_param("SnapshotOptions")
        vpc_options = self._get_param("VPCOptions")
        cognito_options = self._get_param("CognitoOptions")
        encryption_at_rest_options = self._get_param("EncryptionAtRestOptions")
        node_to_node_encryption_options = self._get_param("NodeToNodeEncryptionOptions")
        advanced_options = self._get_param("AdvancedOptions")
        log_publishing_options = self._get_param("LogPublishingOptions")
        domain_endpoint_options = self._get_param("DomainEndpointOptions")
        advanced_security_options = self._get_param("AdvancedSecurityOptions")
        tag_list = self._get_param("TagList")
        auto_tune_options = self._get_param("AutoTuneOptions")
        off_peak_window_options = self._get_param("OffPeakWindowOptions")
        software_update_options = self._get_param("SoftwareUpdateOptions")
        # ElasticSearch specific options
        is_es = self.parsed_url.path.endswith("/es/domain")
        elasticsearch_version = self._get_param("ElasticsearchVersion")
        elasticsearch_cluster_config = self._get_param("ElasticsearchClusterConfig")
        domain = self.opensearch_backend.create_domain(
            domain_name=domain_name,
            engine_version=engine_version,
            cluster_config=cluster_config,
            ebs_options=ebs_options,
            access_policies=access_policies,
            snapshot_options=snapshot_options,
            vpc_options=vpc_options,
            cognito_options=cognito_options,
            encryption_at_rest_options=encryption_at_rest_options,
            node_to_node_encryption_options=node_to_node_encryption_options,
            advanced_options=advanced_options,
            log_publishing_options=log_publishing_options,
            domain_endpoint_options=domain_endpoint_options,
            advanced_security_options=advanced_security_options,
            tag_list=tag_list,
            auto_tune_options=auto_tune_options,
            off_peak_window_options=off_peak_window_options,
            software_update_options=software_update_options,
            is_es=is_es,
            elasticsearch_version=elasticsearch_version,
            elasticsearch_cluster_config=elasticsearch_cluster_config,
        )
        return json.dumps(dict(DomainStatus=domain.to_dict()))

    def get_compatible_versions(self) -> str:
        domain_name = self._get_param("domainName")
        compatible_versions = self.opensearch_backend.get_compatible_versions(
            domain_name=domain_name,
        )
        return json.dumps(dict(CompatibleVersions=compatible_versions))

    def delete_domain(self) -> str:
        domain_name = self.path.split("/")[-1]
        domain = self.opensearch_backend.delete_domain(
            domain_name=domain_name,
        )
        return json.dumps(dict(DomainStatus=domain.to_dict()))

    def describe_domain(self) -> str:
        domain_name = self.path.split("/")[-1]
        if not re.match(r"^[a-z][a-z0-9\-]+$", domain_name):
            raise InvalidDomainName(domain_name)
        domain = self.opensearch_backend.describe_domain(
            domain_name=domain_name,
        )
        return json.dumps(dict(DomainStatus=domain.to_dict()))

    def describe_domain_config(self) -> str:
        domain_name = self.path.split("/")[-1]
        domain = self.opensearch_backend.describe_domain_config(
            domain_name=domain_name,
        )
        return json.dumps(dict(DomainConfig=domain.to_config_dict()))

    def update_domain_config(self) -> str:
        domain_name = self._get_param("DomainName")
        cluster_config = self._get_param("ClusterConfig")
        ebs_options = self._get_param("EBSOptions")
        access_policies = self._get_param("AccessPolicies")
        snapshot_options = self._get_param("SnapshotOptions")
        vpc_options = self._get_param("VPCOptions")
        cognito_options = self._get_param("CognitoOptions")
        encryption_at_rest_options = self._get_param("EncryptionAtRestOptions")
        node_to_node_encryption_options = self._get_param("NodeToNodeEncryptionOptions")
        advanced_options = self._get_param("AdvancedOptions")
        log_publishing_options = self._get_param("LogPublishingOptions")
        domain_endpoint_options = self._get_param("DomainEndpointOptions")
        advanced_security_options = self._get_param("AdvancedSecurityOptions")
        auto_tune_options = self._get_param("AutoTuneOptions")
        off_peak_window_options = self._get_param("OffPeakWindowOptions")
        software_update_options = self._get_param("SoftwareUpdateOptions")
        domain = self.opensearch_backend.update_domain_config(
            domain_name=domain_name,
            cluster_config=cluster_config,
            ebs_options=ebs_options,
            access_policies=access_policies,
            snapshot_options=snapshot_options,
            vpc_options=vpc_options,
            cognito_options=cognito_options,
            encryption_at_rest_options=encryption_at_rest_options,
            node_to_node_encryption_options=node_to_node_encryption_options,
            advanced_options=advanced_options,
            log_publishing_options=log_publishing_options,
            domain_endpoint_options=domain_endpoint_options,
            advanced_security_options=advanced_security_options,
            auto_tune_options=auto_tune_options,
            off_peak_window_options=off_peak_window_options,
            software_update_options=software_update_options,
        )
        return json.dumps(dict(DomainConfig=domain.to_config_dict()))

    def list_tags(self) -> str:
        arn = self._get_param("arn")
        tags = self.opensearch_backend.list_tags(arn)
        return json.dumps({"TagList": tags})

    def add_tags(self) -> str:
        arn = self._get_param("ARN")
        tags = self._get_param("TagList")
        self.opensearch_backend.add_tags(arn, tags)
        return "{}"

    def remove_tags(self) -> str:
        arn = self._get_param("ARN")
        tag_keys = self._get_param("TagKeys")
        self.opensearch_backend.remove_tags(arn, tag_keys)
        return "{}"

    def list_domain_names(self) -> str:
        engine_type = self._get_param("engineType")
        domain_names = self.opensearch_backend.list_domain_names(
            engine_type=engine_type,
        )
        return json.dumps(dict(DomainNames=domain_names))

    def describe_domains(self) -> str:
        domain_names = self._get_param("DomainNames")
        domains = self.opensearch_backend.describe_domains(
            domain_names=domain_names,
        )
        domain_list = [domain.to_dict() for domain in domains]
        return json.dumps({"DomainStatusList": domain_list})



import abc

from moto.stepfunctions.parser.asl.antlr.runtime.ASLParserVisitor import (
    ASLParserVisitor,
)
from moto.stepfunctions.parser.asl.parse.asl_parser import AmazonStateLanguageParser


class StaticAnalyser(ASLParserVisitor, abc.ABC):
    def analyse(self, definition: str) -> None:
        _, parser_rule_context = AmazonStateLanguageParser.parse(definition)
        self.visit(parser_rule_context)



import abc
from typing import Final, Optional

from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class Timeout(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def is_default_value(self) -> bool: ...

    @abc.abstractmethod
    def _eval_seconds(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        seconds = self._eval_seconds(env=env)
        env.stack.append(seconds)


class TimeoutSeconds(Timeout):
    DEFAULT_TIMEOUT_SECONDS: Final[int] = 99999999

    def __init__(self, timeout_seconds: int, is_default: Optional[bool] = None):
        if not isinstance(timeout_seconds, int) and timeout_seconds <= 0:
            raise ValueError(
                f"Expected non-negative integer for TimeoutSeconds, got '{timeout_seconds}' instead."
            )
        self.timeout_seconds: Final[int] = timeout_seconds
        self.is_default: Optional[bool] = is_default

    def is_default_value(self) -> bool:
        if self.is_default is not None:
            return self.is_default
        return self.timeout_seconds == self.DEFAULT_TIMEOUT_SECONDS

    def _eval_seconds(self, env: Environment) -> int:
        return self.timeout_seconds


class TimeoutSecondsJSONata(Timeout):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def is_default_value(self) -> bool:
        return False

    def _eval_seconds(self, env: Environment) -> int:
        self.jsonata_template_value_terminal_expression.eval(env=env)
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        seconds = int(env.stack.pop())
        return seconds


class TimeoutSecondsPath(Timeout):
    def __init__(self, path: str):
        self.path: Final[str] = path

    @classmethod
    def from_raw(cls, path: str):
        return cls(path=path)

    def is_default_value(self) -> bool:
        return False

    def _eval_seconds(self, env: Environment) -> int:
        inp = env.stack[-1]
        seconds = extract_json(self.path, inp)
        if not isinstance(seconds, int) and seconds <= 0:
            raise ValueError(
                f"Expected non-negative integer for TimeoutSecondsPath, got '{seconds}' instead."
            )
        return seconds


class TimeoutSecondsPathVar(TimeoutSecondsPath):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        super().__init__(path=variable_sample.expression)
        self.variable_sample = variable_sample

    def _eval_seconds(self, env: Environment) -> int:
        self.variable_sample.eval(env=env)
        seconds = env.stack.pop()
        if not isinstance(seconds, int) and seconds <= 0:
            raise ValueError(
                f"Expected non-negative integer for TimeoutSecondsPath, got '{seconds}' instead."
            )
        return seconds


class EvalTimeoutError(TimeoutError):
    pass



from collections import OrderedDict
from typing import Dict

from localstack.services.stores import AccountRegionBundle, BaseStore, LocalAttribute

from moto.stepfunctions.parser.api import Arn
from moto.stepfunctions.parser.backend.activity import Activity
from moto.stepfunctions.parser.backend.execution import Execution
from moto.stepfunctions.parser.backend.state_machine import StateMachineInstance


class SFNStore(BaseStore):
    # Maps ARNs to state machines.
    state_machines: Dict[Arn, StateMachineInstance] = LocalAttribute(default=dict)
    # Maps Execution-ARNs to state machines.
    executions: Dict[Arn, Execution] = LocalAttribute(
        default=OrderedDict
    )  # TODO: when snapshot to pods stop execution(?)
    activities: OrderedDict[Arn, Activity] = LocalAttribute(default=dict)


sfn_stores: AccountRegionBundle = AccountRegionBundle("stepfunctions", SFNStore)



# Sourced from https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cfn-lambda-function-code-cfnresponsemodule.html
# 01/Nov/2021

# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: MIT-0

from __future__ import print_function

import json
from typing import Any

import urllib3

SUCCESS = "SUCCESS"
FAILED = "FAILED"

http = urllib3.PoolManager()


def send(
    event: Any,
    context: Any,
    responseStatus: Any,
    responseData: Any,
    physicalResourceId: Any = None,
    noEcho: bool = False,
    reason: Any = None,
) -> None:
    responseUrl = event["ResponseURL"]

    print(responseUrl)  # noqa: T201

    responseBody = {
        "Status": responseStatus,
        "Reason": reason
        or "See the details in CloudWatch Log Stream: {}".format(
            context.log_stream_name
        ),
        "PhysicalResourceId": physicalResourceId or context.log_stream_name,
        "StackId": event["StackId"],
        "RequestId": event["RequestId"],
        "LogicalResourceId": event["LogicalResourceId"],
        "NoEcho": noEcho,
        "Data": responseData,
    }

    json_responseBody = json.dumps(responseBody)

    print("Response body:")  # noqa: T201
    print(json_responseBody)  # noqa: T201

    headers = {"content-type": "", "content-length": str(len(json_responseBody))}

    try:
        response = http.request(  # type: ignore
            "PUT", responseUrl, headers=headers, body=json_responseBody
        )
        print("Status code:", response.status)  # noqa: T201

    except Exception as e:
        print("send(..) failed executing http.request(..):", e)  # noqa: T201



"""Exceptions raised by the quicksight service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("ResourceNotFoundException", msg)


class InvalidParameterValueException(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("InvalidParameterValueException", msg)


class ValidationException(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("ValidationException", msg)


class SchemaException(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("SchemaException", msg)


class ParamValidationError(JsonRESTError):
    def __init__(self, msg: str):
        super().__init__("ParamValidationError", msg)



import boto3
import pytest
from botocore.errorfactory import ClientError

from moto import mock_aws

from . import cloudfront_test_scaffolding as scaffold


@mock_aws
def test_create_invalidation_with_single_path():
    client = boto3.client("cloudfront", region_name="us-west-1")
    config = scaffold.example_distribution_config("ref")
    resp = client.create_distribution(DistributionConfig=config)
    dist_id = resp["Distribution"]["Id"]

    resp = client.create_invalidation(
        DistributionId=dist_id,
        InvalidationBatch={
            "Paths": {"Quantity": 1, "Items": ["/path1"]},
            "CallerReference": "ref2",
        },
    )

    assert "Location" in resp
    assert "Invalidation" in resp

    assert "Id" in resp["Invalidation"]
    assert resp["Invalidation"]["Status"] == "COMPLETED"
    assert "CreateTime" in resp["Invalidation"]
    assert resp["Invalidation"]["InvalidationBatch"] == {
        "Paths": {"Quantity": 1, "Items": ["/path1"]},
        "CallerReference": "ref2",
    }


@mock_aws
def test_create_invalidation_with_multiple_paths():
    client = boto3.client("cloudfront", region_name="us-west-1")
    config = scaffold.example_distribution_config("ref")
    resp = client.create_distribution(DistributionConfig=config)
    dist_id = resp["Distribution"]["Id"]

    resp = client.create_invalidation(
        DistributionId=dist_id,
        InvalidationBatch={
            "Paths": {"Quantity": 2, "Items": ["/path1", "/path2"]},
            "CallerReference": "ref2",
        },
    )

    assert "Location" in resp
    assert "Invalidation" in resp

    assert "Id" in resp["Invalidation"]
    assert resp["Invalidation"]["Status"] == "COMPLETED"
    assert "CreateTime" in resp["Invalidation"]
    assert resp["Invalidation"]["InvalidationBatch"] == {
        "Paths": {"Quantity": 2, "Items": ["/path1", "/path2"]},
        "CallerReference": "ref2",
    }


@mock_aws
def test_list_invalidations():
    client = boto3.client("cloudfront", region_name="us-west-1")
    config = scaffold.example_distribution_config("ref")
    resp = client.create_distribution(DistributionConfig=config)
    dist_id = resp["Distribution"]["Id"]
    resp = client.create_invalidation(
        DistributionId=dist_id,
        InvalidationBatch={
            "Paths": {"Quantity": 2, "Items": ["/path1", "/path2"]},
            "CallerReference": "ref2",
        },
    )
    invalidation_id = resp["Invalidation"]["Id"]

    resp = client.list_invalidations(DistributionId=dist_id)

    assert "NextMarker" not in resp["InvalidationList"]
    assert resp["InvalidationList"]["MaxItems"] == 100
    assert resp["InvalidationList"]["IsTruncated"] is False
    assert resp["InvalidationList"]["Quantity"] == 1
    assert len(resp["InvalidationList"]["Items"]) == 1
    assert resp["InvalidationList"]["Items"][0]["Id"] == invalidation_id
    assert "CreateTime" in resp["InvalidationList"]["Items"][0]
    assert resp["InvalidationList"]["Items"][0]["Status"] == "COMPLETED"


@mock_aws
def test_list_invalidations__no_entries():
    client = boto3.client("cloudfront", region_name="us-west-1")

    resp = client.list_invalidations(DistributionId="SPAM")

    assert "NextMarker" not in resp["InvalidationList"]
    assert resp["InvalidationList"]["MaxItems"] == 100
    assert resp["InvalidationList"]["IsTruncated"] is False
    assert resp["InvalidationList"]["Quantity"] == 0
    assert "Items" not in resp["InvalidationList"]


@mock_aws
def test_get_invalidation():
    client = boto3.client("cloudfront", region_name="us-west-1")
    config = scaffold.example_distribution_config("ref")
    resp = client.create_distribution(DistributionConfig=config)
    dist_id = resp["Distribution"]["Id"]

    createResp = client.create_invalidation(
        DistributionId=dist_id,
        InvalidationBatch={
            "Paths": {"Quantity": 1, "Items": ["/path1"]},
            "CallerReference": "ref2",
        },
    )
    existed = createResp["Invalidation"]
    resp = client.get_invalidation(DistributionId=dist_id, Id=existed["Id"])
    assert "Invalidation" in resp
    returned = resp["Invalidation"]
    assert returned["Id"] == existed["Id"]
    assert returned["Status"] == existed["Status"]
    assert returned["CreateTime"] == existed["CreateTime"]
    assert returned["InvalidationBatch"] == existed["InvalidationBatch"]


@mock_aws
def test_get_invalidation_dist_not_found():
    client = boto3.client("cloudfront", region_name="us-west-1")

    with pytest.raises(ClientError) as exc:
        client.get_invalidation(DistributionId="notfound", Id="notfound")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchDistribution"
    assert err["Message"] == "The specified distribution does not exist."


@mock_aws
def test_get_invalidation_id_not_found():
    client = boto3.client("cloudfront", region_name="us-west-1")
    config = scaffold.example_distribution_config("ref")
    resp = client.create_distribution(DistributionConfig=config)
    dist_id = resp["Distribution"]["Id"]

    with pytest.raises(ClientError) as exc:
        client.get_invalidation(DistributionId=dist_id, Id="notfound")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchInvalidation"
    assert err["Message"] == "The specified invalidation does not exist."



import fnmatch
from typing import Any

from moto.stepfunctions.parser.asl.component.state.choice.comparison.comparison_operator_type import (
    ComparisonOperatorType,
)
from moto.stepfunctions.parser.asl.component.state.choice.comparison.operator.operator import (
    Operator,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class StringEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = variable == comparison_value
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringEquals._compare(variable, value)
        env.stack.append(res)


class StringEqualsPath(StringEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = StringEqualsPath._compare(variable, comp_value)
        env.stack.append(res)


class StringGreaterThan(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringGreaterThan)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = variable > comparison_value
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringGreaterThan._compare(variable, value)
        env.stack.append(res)


class StringGreaterThanPath(StringGreaterThan):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringGreaterThanPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = StringGreaterThanPath._compare(variable, comp_value)
        env.stack.append(res)


class StringGreaterThanEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringGreaterThanEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = variable >= comparison_value
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringGreaterThanEquals._compare(variable, value)
        env.stack.append(res)


class StringGreaterThanEqualsPath(StringGreaterThanEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringGreaterThanEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = StringGreaterThanEqualsPath._compare(variable, comp_value)
        env.stack.append(res)


class StringLessThan(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringLessThan)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = variable < comparison_value
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringLessThan._compare(variable, value)
        env.stack.append(res)


class StringLessThanPath(StringLessThan):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringLessThanPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = StringLessThanPath._compare(variable, comp_value)
        env.stack.append(res)


class StringLessThanEquals(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringLessThanEquals)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = variable <= comparison_value
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringLessThanEquals._compare(variable, value)
        env.stack.append(res)


class StringLessThanEqualsPath(StringLessThanEquals):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringLessThanEqualsPath)

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        inp = env.stack[-1]
        comp_value = extract_json(value, inp)
        res = StringLessThanEqualsPath._compare(variable, comp_value)
        env.stack.append(res)


class StringMatches(Operator):
    @staticmethod
    def impl_name() -> str:
        return str(ComparisonOperatorType.StringMatches)

    @staticmethod
    def _compare(variable: Any, comparison_value: Any) -> bool:
        res = False
        if isinstance(variable, str):
            res = fnmatch.fnmatch(variable, comparison_value)
        return res

    @staticmethod
    def eval(env: Environment, value: Any) -> None:
        variable = env.stack.pop()
        res = StringMatches._compare(variable, value)
        env.stack.append(res)



import json

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import (
    ActivityDoesNotExist,
    ActivityFailedEventDetails,
    ActivityScheduledEventDetails,
    ActivityStartedEventDetails,
    ActivitySucceededEventDetails,
    ActivityTimedOutEventDetails,
    ExecutionFailedEventDetails,
    HistoryEventExecutionDataDetails,
    HistoryEventType,
)
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.timeouts.timeout import (
    EvalTimeoutError,
    TimeoutSeconds,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ActivityResource,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.state_task import (
    StateTask,
)
from moto.stepfunctions.parser.asl.eval.callback.callback import (
    ActivityTaskStartOutcome,
    CallbackOutcomeFailure,
    CallbackOutcomeFailureError,
    CallbackOutcomeSuccess,
    CallbackTimeoutError,
    HeartbeatTimeoutError,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str


class StateTaskActivity(StateTask):
    resource: ActivityResource

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, TimeoutError):
            return FailureEvent(
                env=env,
                error_name=StatesErrorName(typ=StatesErrorNameType.StatesTimeout),
                event_type=HistoryEventType.ActivityTimedOut,
                event_details=EventDetails(
                    activityTimedOutEventDetails=ActivityTimedOutEventDetails(
                        error=StatesErrorNameType.StatesTimeout.to_name(),
                    )
                ),
            )

        if isinstance(ex, FailureEventException):
            raise ex

        if isinstance(ex, CallbackOutcomeFailureError):
            error = ex.callback_outcome_failure.error
            error_name = CustomErrorName(error)
            cause = ex.callback_outcome_failure.cause
        else:
            error_name = StatesErrorName(typ=StatesErrorNameType.StatesRuntime)
            error = error_name.error_name
            cause = (
                ex.response["Error"]["Message"]
                if isinstance(ex, ClientError)
                else str(ex)
            )
        return FailureEvent(
            env=env,
            error_name=error_name,
            event_type=HistoryEventType.ActivityFailed,
            event_details=EventDetails(
                activityFailedEventDetails=ActivityFailedEventDetails(
                    error=error, cause=cause
                )
            ),
        )

    def _eval_parameters(self, env: Environment) -> dict:
        if self.parargs:
            self.parargs.eval(env=env)
        activity_input = env.stack.pop()
        return activity_input

    def _eval_execution(self, env: Environment) -> None:
        # Compute the task input.
        activity_task_input = self._eval_parameters(env=env)
        if not isinstance(activity_task_input, str):
            activity_task_input = to_json_str(activity_task_input)

        # Compute the timeout and heartbeat for this task.
        timeout_seconds = TimeoutSeconds.DEFAULT_TIMEOUT_SECONDS

        if not self.timeout.is_default_value():
            self.timeout.eval(env=env)
            timeout_seconds = env.stack.pop()

        heartbeat_seconds = None
        if self.heartbeat:
            self.heartbeat.eval(env=env)
            heartbeat_seconds = env.stack.pop()

        # Publish the activity task on the callback manager.
        task_token = env.states.context_object.update_task_token()
        try:
            callback_endpoint = env.callback_pool_manager.add_activity_task(
                callback_id=task_token,
                activity_arn=self.resource.resource_arn,
                activity_input=activity_task_input,
            )
        except ActivityDoesNotExist:
            failure_event = FailureEvent(
                env=env,
                error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                event_type=HistoryEventType.ExecutionFailed,
                event_details=EventDetails(
                    executionFailedEventDetails=ExecutionFailedEventDetails(
                        error=StatesErrorNameType.StatesRuntime.to_name(),
                        cause="The activity activity_arn does not exist.",
                    )
                ),
            )
            raise FailureEventException(failure_event=failure_event)

        # Log the task is scheduled.
        scheduled_event_details = ActivityScheduledEventDetails(
            resource=self.resource.resource_arn,
            input=activity_task_input,
            inputDetails=HistoryEventExecutionDataDetails(
                truncated=False  # Always False for api calls.
            ),
        )
        if timeout_seconds != TimeoutSeconds.DEFAULT_TIMEOUT_SECONDS:
            scheduled_event_details["timeoutInSeconds"] = timeout_seconds
        if heartbeat_seconds is not None:
            scheduled_event_details["heartbeatInSeconds"] = heartbeat_seconds
        env.event_manager.add_event(
            context=env.event_history_context,
            event_type=HistoryEventType.ActivityScheduled,
            event_details=EventDetails(
                activityScheduledEventDetails=scheduled_event_details
            ),
        )

        # Await for the task to be sampled with timeout.
        activity_task_start_endpoint = (
            callback_endpoint.get_activity_task_start_endpoint()
        )
        task_start_outcome = activity_task_start_endpoint.wait(
            timeout_seconds=timeout_seconds
        )
        # Log the task was sampled or timeout error if not.
        if isinstance(task_start_outcome, ActivityTaskStartOutcome):
            started_event_details = ActivityStartedEventDetails()
            if task_start_outcome.worker_name is not None:
                started_event_details["workerName"] = task_start_outcome.worker_name
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=HistoryEventType.ActivityStarted,
                event_details=EventDetails(
                    activityStartedEventDetails=started_event_details
                ),
            )
        else:
            raise EvalTimeoutError()

        # Await for the task outcome, with a heartbeat or timeout strategy.
        outcome = None
        if heartbeat_seconds is None:
            # Total timeout is already handled upstream. Here we specify a timeout to allow this child operation to
            # terminate gracefully sooner. This is why we don't compute the residual outcome.
            outcome = callback_endpoint.wait(timeout=timeout_seconds)
        else:
            heartbeat_endpoint = callback_endpoint.setup_heartbeat_endpoint(
                heartbeat_seconds=heartbeat_seconds
            )
            while (
                env.is_running() and outcome is None
            ):  # Until subprocess hasn't timed out or result wasn't received.
                received = heartbeat_endpoint.clear_and_wait()
                if not received and env.is_running():  # Heartbeat timed out.
                    raise HeartbeatTimeoutError()
                outcome = callback_endpoint.get_outcome()

        if outcome is None:
            raise CallbackTimeoutError()
        if isinstance(outcome, CallbackOutcomeSuccess):
            outcome_output = json.loads(outcome.output)
            env.stack.append(outcome_output)
        elif isinstance(outcome, CallbackOutcomeFailure):
            raise CallbackOutcomeFailureError(callback_outcome_failure=outcome)
        else:
            raise NotImplementedError(
                f"Unsupported CallbackOutcome type '{type(outcome)}'."
            )

        env.event_manager.add_event(
            context=env.event_history_context,
            event_type=HistoryEventType.ActivitySucceeded,
            event_details=EventDetails(
                activitySucceededEventDetails=ActivitySucceededEventDetails(
                    output=outcome.output,
                    outputDetails=HistoryEventExecutionDataDetails(
                        truncated=False  # Always False for api calls.
                    ),
                )
            ),
        )



import abc
from typing import Final

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.jsonata.jsonata_template_value_terminal import (
    JSONataTemplateValueTerminalExpression,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.asl.utils.json_path import extract_json

DEFAULT_MAX_CONCURRENCY_VALUE: int = 0  # No limit.


class MaxConcurrencyDecl(EvalComponent, abc.ABC):
    @abc.abstractmethod
    def _eval_max_concurrency(self, env: Environment) -> int: ...

    def _eval_body(self, env: Environment) -> None:
        max_concurrency_value = self._eval_max_concurrency(env=env)
        env.stack.append(max_concurrency_value)


class MaxConcurrency(MaxConcurrencyDecl):
    max_concurrency_value: int

    def __init__(self, num: int = DEFAULT_MAX_CONCURRENCY_VALUE):
        super().__init__()
        self.max_concurrency_value = num

    def _eval_max_concurrency(self, env: Environment) -> int:
        return self.max_concurrency_value


class MaxConcurrencyJSONata(MaxConcurrencyDecl):
    jsonata_template_value_terminal_expression: Final[
        JSONataTemplateValueTerminalExpression
    ]

    def __init__(
        self,
        jsonata_template_value_terminal_expression: JSONataTemplateValueTerminalExpression,
    ):
        super().__init__()
        self.jsonata_template_value_terminal_expression = (
            jsonata_template_value_terminal_expression
        )

    def _eval_max_concurrency(self, env: Environment) -> int:
        self.jsonata_template_value_terminal_expression.eval(env=env)
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        seconds = int(env.stack.pop())
        return seconds


class MaxConcurrencyPathVar(MaxConcurrency):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        super().__init__()
        self.variable_sample = variable_sample

    def _eval_max_concurrency(self, env: Environment) -> int:
        self.variable_sample.eval(env=env)
        # TODO: add snapshot tests to verify AWS's behaviour about non integer values.
        max_concurrency: int = int(env.stack.pop())
        return max_concurrency


class MaxConcurrencyPath(MaxConcurrency):
    max_concurrency_path: str

    def __init__(self, max_concurrency_path: str):
        super().__init__()
        self.max_concurrency_path = max_concurrency_path

    def _eval_max_concurrency(self, env: Environment) -> int:
        inp = env.stack[-1]
        max_concurrency_value = extract_json(self.max_concurrency_path, inp)

        error_cause = None
        if not isinstance(max_concurrency_value, int):
            value_str = (
                to_json_str(max_concurrency_value)
                if not isinstance(max_concurrency_value, str)
                else max_concurrency_value
            )
            error_cause = f'The MaxConcurrencyPath field refers to value "{value_str}" which is not a valid integer: {self.max_concurrency_path}'
        elif max_concurrency_value < 0:
            error_cause = f"Expected non-negative integer for MaxConcurrency, got '{max_concurrency_value}' instead."

        if error_cause is not None:
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=StatesErrorNameType.StatesRuntime.to_name(),
                            cause=error_cause,
                        )
                    ),
                )
            )

        return max_concurrency_value



import logging
import threading

from moto.stepfunctions.parser.api import (
    ExecutionFailedEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.state.state import CommonStateField
from moto.stepfunctions.parser.asl.eval.test_state.environment import (
    TestStateEnvironment,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.utils import TMP_THREADS

LOG = logging.getLogger(__name__)

TEST_CASE_EXECUTION_TIMEOUT_SECONDS: int = 300  # 5 minutes.


class TestStateProgram(EvalComponent):
    test_state: CommonStateField

    def __init__(
        self,
        test_state: CommonStateField,
    ):
        self.test_state = test_state

    def eval(self, env: TestStateEnvironment) -> None:
        env.next_state_name = self.test_state.name
        worker_thread = threading.Thread(target=super().eval, args=(env,), daemon=True)
        TMP_THREADS.append(worker_thread)
        worker_thread.start()
        worker_thread.join(timeout=TEST_CASE_EXECUTION_TIMEOUT_SECONDS)
        is_timeout = worker_thread.is_alive()
        if is_timeout:
            env.set_timed_out()

    def _eval_body(self, env: TestStateEnvironment) -> None:
        try:
            env.inspection_data["input"] = to_json_str(env.states.get_input())
            self.test_state.eval(env=env)
        except FailureEventException as ex:
            env.set_error(error=ex.get_execution_failed_event_details())
        except Exception as ex:
            cause = f"{type(ex).__name__}({str(ex)})"
            LOG.error("Stepfunctions computation ended with exception '%s'.", cause)
            env.set_error(
                ExecutionFailedEventDetails(
                    error=StatesErrorName(
                        typ=StatesErrorNameType.StatesRuntime
                    ).error_name,
                    cause=cause,
                )
            )



from typing import Any, Final, List

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_string import (
    FunctionArgumentString,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.states_function import (
    StatesFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class StatesFunctionFormat(StatesFunction):
    _DELIMITER: Final[str] = "{}"

    def __init__(self, arg_list: FunctionArgumentList):
        super().__init__(
            states_name=StatesFunctionName(function_type=StatesFunctionNameType.Format),
            arg_list=arg_list,
        )
        if arg_list.size > 0:
            raise ValueError(
                f"Expected at least 1 argument for function type '{type(self)}', but got: '{arg_list}'."
            )
        if not isinstance(arg_list.arg_list[0], FunctionArgumentString):
            raise ValueError(
                f"Expected the first argument for function type '{type(self)}' to be a string, but got: '{arg_list.arg_list[0]}'."
            )

    def _eval_body(self, env: Environment) -> None:
        # TODO: investigate behaviour for incorrect number of arguments in string format.
        self.arg_list.eval(env=env)

        values: List[Any] = list()
        for _ in range(self.arg_list.size):
            values.append(env.stack.pop())
        string_format: str = values.pop()
        values.reverse()

        string_format_parts: List[str] = string_format.split(self._DELIMITER)
        string_result: str = ""
        for part in string_format_parts:
            string_result += part
            string_result += values.pop()

        env.stack.append(string_result)



from typing import Any

from moto.core.exceptions import RESTError

EXCEPTION_RESPONSE = """<?xml version="1.0"?>
<ErrorResponse xmlns="http://elasticache.amazonaws.com/doc/2015-02-02/">
  <Error>
    <Type>Sender</Type>
    <Code>{{ error_type }}</Code>
    <Message>{{ message }}</Message>
  </Error>
  <{{ request_id_tag }}>30c0dedb-92b1-4e2b-9be4-1188e3ed86ab</{{ request_id_tag }}>
</ErrorResponse>"""


class ElastiCacheException(RESTError):
    code = 400
    extended_templates = {"ecerror": EXCEPTION_RESPONSE}
    env = RESTError.extended_environment(extended_templates)

    def __init__(self, code: str, message: str, **kwargs: Any):
        kwargs.setdefault("template", "ecerror")
        super().__init__(code, message)


class PasswordTooShort(ElastiCacheException):
    code = 404

    def __init__(self) -> None:
        super().__init__(
            "InvalidParameterValue",
            message="Passwords length must be between 16-128 characters.",
        )


class PasswordRequired(ElastiCacheException):
    code = 404

    def __init__(self) -> None:
        super().__init__(
            "InvalidParameterValue",
            message="No password was provided. If you want to create/update the user without password, please use the NoPasswordRequired flag.",
        )


class InvalidParameterValueException(ElastiCacheException):
    code = 404

    def __init__(self, message: str) -> None:
        super().__init__(
            "InvalidParameterValue",
            message=message,
        )


class InvalidParameterCombinationException(ElastiCacheException):
    code = 404

    def __init__(self, message: str) -> None:
        super().__init__(
            "InvalidParameterCombination",
            message=message,
        )


class UserAlreadyExists(ElastiCacheException):
    code = 404

    def __init__(self) -> None:
        super().__init__("UserAlreadyExists", message="User user1 already exists.")


class UserNotFound(ElastiCacheException):
    code = 404

    def __init__(self, user_id: str):
        super().__init__("UserNotFound", message=f"User {user_id} not found.")


class CacheClusterAlreadyExists(ElastiCacheException):
    code = 404

    def __init__(self, cache_cluster_id: str):
        (
            super().__init__(
                "CacheClusterAlreadyExists",
                message=f"Cache cluster {cache_cluster_id} already exists.",
            ),
        )


class CacheClusterNotFound(ElastiCacheException):
    code = 404

    def __init__(self, cache_cluster_id: str):
        super().__init__(
            "CacheClusterNotFound",
            message=f"Cache cluster {cache_cluster_id} not found.",
        )


class InvalidARNFault(ElastiCacheException):
    code = 400

    def __init__(self, arn: str):
        super().__init__(
            "InvalidARNFault",
            message=f"ARN {arn} is invalid.",
        )



from typing import Optional

from moto.core.exceptions import JsonRESTError


class DataSyncClientError(JsonRESTError):
    code = 400


class InvalidRequestException(DataSyncClientError):
    def __init__(self, msg: Optional[str] = None):
        self.code = 400
        super().__init__("InvalidRequestException", msg or "The request is not valid.")



"""Handles incoming osis requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import OpenSearchIngestionBackend, osis_backends


class OpenSearchIngestionResponse(BaseResponse):
    """Handler for OpenSearchIngestion requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="osis")

    @property
    def osis_backend(self) -> OpenSearchIngestionBackend:
        """Return backend instance specific for this region."""
        return osis_backends[self.current_account][self.region]

    def create_pipeline(self) -> str:
        params = json.loads(self.body)
        pipeline_name = params.get("PipelineName")
        min_units = params.get("MinUnits")
        max_units = params.get("MaxUnits")
        pipeline_configuration_body = params.get("PipelineConfigurationBody")
        log_publishing_options = params.get("LogPublishingOptions")
        vpc_options = params.get("VpcOptions")
        buffer_options = params.get("BufferOptions")
        encryption_at_rest_options = params.get("EncryptionAtRestOptions")
        tags = params.get("Tags")
        pipeline = self.osis_backend.create_pipeline(
            pipeline_name=pipeline_name,
            min_units=min_units,
            max_units=max_units,
            pipeline_configuration_body=pipeline_configuration_body,
            log_publishing_options=log_publishing_options,
            vpc_options=vpc_options,
            buffer_options=buffer_options,
            encryption_at_rest_options=encryption_at_rest_options,
            tags=tags,
        )
        return json.dumps(dict(Pipeline=pipeline.to_dict()))

    def delete_pipeline(self) -> str:
        pipeline_name = self._get_param("PipelineName")
        self.osis_backend.delete_pipeline(
            pipeline_name=pipeline_name,
        )
        return json.dumps(dict())

    def get_pipeline(self) -> str:
        pipeline_name = self._get_param("PipelineName")
        pipeline = self.osis_backend.get_pipeline(
            pipeline_name=pipeline_name,
        )
        return json.dumps(dict(Pipeline=pipeline.to_dict()))

    def list_pipelines(self) -> str:
        max_results = self._get_int_param("MaxResults")
        next_token = self._get_param("NextToken")
        pipelines, next_token = self.osis_backend.list_pipelines(
            max_results=max_results,
            next_token=next_token,
        )
        return json.dumps(
            dict(nextToken=next_token, Pipelines=[p.to_short_dict() for p in pipelines])
        )

    def list_tags_for_resource(self) -> str:
        arn = self._get_param("arn")
        tags = self.osis_backend.list_tags_for_resource(arn=arn)
        return json.dumps(dict(tags))

    def update_pipeline(self) -> str:
        params = json.loads(self.body)
        pipeline_name = self.path.split("/")[-1]
        min_units = params.get("MinUnits")
        max_units = params.get("MaxUnits")
        pipeline_configuration_body = params.get("PipelineConfigurationBody")
        log_publishing_options = params.get("LogPublishingOptions")
        buffer_options = params.get("BufferOptions")
        encryption_at_rest_options = params.get("EncryptionAtRestOptions")
        pipeline = self.osis_backend.update_pipeline(
            pipeline_name=pipeline_name,
            min_units=min_units,
            max_units=max_units,
            pipeline_configuration_body=pipeline_configuration_body,
            log_publishing_options=log_publishing_options,
            buffer_options=buffer_options,
            encryption_at_rest_options=encryption_at_rest_options,
        )
        # TODO: adjust response
        return json.dumps(dict(Pipeline=pipeline.to_dict()))

    def tag_resource(self) -> str:
        params = json.loads(self.body)
        arn = self._get_param("arn")
        tags = params.get("Tags")
        self.osis_backend.tag_resource(
            arn=arn,
            tags=tags,
        )
        return json.dumps(dict())

    def untag_resource(self) -> str:
        params = json.loads(self.body)
        arn = self._get_param("arn")
        tag_keys = params.get("TagKeys")
        self.osis_backend.untag_resource(
            arn=arn,
            tag_keys=tag_keys,
        )
        return json.dumps(dict())

    def start_pipeline(self) -> str:
        pipeline_name = self._get_param("PipelineName")
        pipeline = self.osis_backend.start_pipeline(
            pipeline_name=pipeline_name,
        )
        return json.dumps(dict(Pipeline=pipeline.to_dict()))

    def stop_pipeline(self) -> str:
        pipeline_name = self._get_param("PipelineName")
        pipeline = self.osis_backend.stop_pipeline(
            pipeline_name=pipeline_name,
        )
        return json.dumps(dict(Pipeline=pipeline.to_dict()))
